"""
FastViT, adapted from
https://github.com/apple/ml-fastvit/blob/main/models/fastvit.py
and
https://github.com/apple/ml-mobileclip/blob/main/mobileclip/models/mci.py

Paper "FastViT: A Fast Hybrid Vision Transformer using Structural Reparameterization",
https://arxiv.org/abs/2303.14189
and
Paper "MobileCLIP: Fast Image-Text Models through Multi-Modal Reinforced Training",
https://arxiv.org/abs/2311.17049

Changes from original:
* Fixed ReparamLargeKernelConv activation (at forward)
"""

# Reference license: Apple MIT License (both)

from collections import OrderedDict
from typing import Any
from typing import Literal
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import SqueezeExcitation
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_tensor
from birder.layers import LayerScale2d
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenRetentionResultType
from birder.net.mobileone import MobileOneBlock


class ReparamLargeKernelConv(nn.Module):
    """
    Building Block of RepLKNet

    This class defines overparameterized large kernel conv block
    introduced in RepLKNet https://arxiv.org/abs/2203.06717

    Reference: https://github.com/DingXiaoH/RepLKNet-pytorch (MIT License)
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int,
        groups: int,
        small_kernel: int,
        use_se: bool,
        reparameterized: bool,
    ) -> None:
        super().__init__()
        assert small_kernel <= kernel_size

        self.reparameterized = reparameterized
        self.stride = stride
        self.groups = groups
        self.in_channels = in_channels
        self.out_channels = out_channels

        self.kernel_size = kernel_size
        self.small_kernel = small_kernel
        self.padding = kernel_size // 2
        if reparameterized is True:
            self.lkb_reparam = nn.Conv2d(
                in_channels,
                out_channels,
                kernel_size=kernel_size,
                stride=stride,
                padding=self.padding,
                groups=self.groups,
                bias=True,
            )
        else:
            self.lkb_reparam = None

            self.lkb_origin = nn.Sequential()
            self.lkb_origin.add_module(
                "conv",
                nn.Conv2d(
                    in_channels,
                    out_channels,
                    kernel_size=kernel_size,
                    stride=self.stride,
                    padding=self.padding,
                    groups=self.groups,
                    bias=False,
                ),
            )
            self.lkb_origin.add_module("bn", nn.BatchNorm2d(num_features=out_channels))

            self.small_conv = nn.Sequential()
            self.small_conv.add_module(
                "conv",
                nn.Conv2d(
                    in_channels,
                    out_channels,
                    kernel_size=small_kernel,
                    stride=self.stride,
                    padding=small_kernel // 2,
                    groups=self.groups,
                    bias=False,
                ),
            )
            self.small_conv.add_module("bn", nn.BatchNorm2d(num_features=out_channels))

        if use_se is True:
            self.se = SqueezeExcitation(out_channels, out_channels // 4)
        else:
            self.se = nn.Identity()

        self.activation = nn.GELU()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.lkb_reparam is not None:
            # Reparameterized forward pass
            x = self.lkb_reparam(x)
        else:
            # Multi-branched train-time forward pass
            x = self.lkb_origin(x) + self.small_conv(x)

        x = self.se(x)
        x = self.activation(x)

        return x

    def reparameterize(self) -> None:
        if self.reparameterized is True:
            return

        (eq_k, eq_b) = self._get_kernel_bias()
        self.lkb_reparam = nn.Conv2d(
            self.in_channels,
            self.out_channels,
            kernel_size=self.kernel_size,
            stride=self.stride,
            padding=self.padding,
            groups=self.groups,
            bias=True,
        )

        self.lkb_reparam.weight.data = eq_k
        self.lkb_reparam.bias.data = eq_b

        # Delete un-used branches
        for param in self.parameters():
            param.detach_()

        del self.lkb_origin
        del self.small_conv

        self.reparameterized = True

    def _get_kernel_bias(self) -> tuple[torch.Tensor, torch.Tensor]:
        (eq_k, eq_b) = self._fuse_bn_tensor(self.lkb_origin.conv, self.lkb_origin.bn)

        (small_k, small_b) = self._fuse_bn_tensor(self.small_conv.conv, self.small_conv.bn)
        eq_b += small_b
        eq_k += F.pad(small_k, [(self.kernel_size - self.small_kernel) // 2] * 4)

        return (eq_k, eq_b)

    @staticmethod
    def _fuse_bn_tensor(conv: torch.Tensor, bn: nn.BatchNorm2d) -> tuple[torch.Tensor, torch.Tensor]:
        kernel_value = conv.weight
        running_mean = bn.running_mean
        running_var = bn.running_var
        gamma = bn.weight
        beta = bn.bias
        eps = bn.eps

        std = (running_var + eps).sqrt()
        t = (gamma / std).reshape(-1, 1, 1, 1)

        return (kernel_value * t, beta - running_mean * gamma / std)


class MHSA(nn.Module):
    def __init__(
        self,
        dim: int,
        head_dim: int,
        qkv_bias: bool,
        attn_drop: float,
        proj_drop: float,
    ) -> None:
        super().__init__()
        assert dim % head_dim == 0, "dim must be divisible by head_dim"

        self.head_dim = head_dim
        self.num_heads = dim // head_dim
        self.scale = head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, C, H, W) = x.shape
        N = H * W
        x = x.flatten(2).transpose(-2, -1)  # (B, N, C)
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, self.head_dim).permute(2, 0, 3, 1, 4)
        (q, k, v) = qkv.unbind(0)
        x = F.scaled_dot_product_attention(  # pylint: disable=not-callable
            q, k, v, dropout_p=self.attn_drop.p if self.training else 0.0, scale=self.scale
        )

        x = x.transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        x = x.transpose(-2, -1).reshape(B, C, H, W)

        return x


class PatchEmbed(nn.Module):
    def __init__(
        self,
        patch_size: int,
        stride: int,
        in_channels: int,
        embed_dim: int,
        use_se: bool,
        reparameterized: bool,
    ) -> None:
        super().__init__()
        self.proj = nn.Sequential(
            ReparamLargeKernelConv(
                in_channels=in_channels,
                out_channels=embed_dim,
                kernel_size=patch_size,
                stride=stride,
                groups=in_channels,
                small_kernel=3,
                use_se=use_se,
                reparameterized=reparameterized,
            ),
            MobileOneBlock(
                in_channels=embed_dim,
                out_channels=embed_dim,
                kernel_size=1,
                stride=1,
                padding=0,
                groups=1,
                use_se=False,
                use_act=True,
                use_scale_branch=True,
                num_conv_branches=1,
                reparameterized=reparameterized,
                activation_layer=nn.GELU,
            ),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.proj(x)


class RepMixer(nn.Module):
    def __init__(
        self,
        dim: int,
        kernel_size: int,
        layer_scale_init_value: Optional[float],
        reparameterized: bool,
    ) -> None:
        super().__init__()
        self.dim = dim
        self.kernel_size = kernel_size
        self.reparameterized = reparameterized

        if reparameterized is True:
            self.reparam_conv = nn.Conv2d(
                self.dim,
                self.dim,
                kernel_size=self.kernel_size,
                stride=(1, 1),
                padding=self.kernel_size // 2,
                groups=self.dim,
                bias=True,
            )
        else:
            self.reparam_conv = None

            self.norm = MobileOneBlock(
                in_channels=dim,
                out_channels=dim,
                kernel_size=kernel_size,
                stride=1,
                padding=kernel_size // 2,
                groups=dim,
                use_se=False,
                use_act=False,
                use_scale_branch=False,
                num_conv_branches=0,
                reparameterized=False,
                activation_layer=nn.GELU,
            )
            self.mixer = MobileOneBlock(
                in_channels=dim,
                out_channels=dim,
                kernel_size=kernel_size,
                stride=1,
                padding=kernel_size // 2,
                groups=dim,
                use_se=False,
                use_act=False,
                use_scale_branch=True,
                num_conv_branches=1,
                reparameterized=False,
                activation_layer=nn.GELU,
            )
            if layer_scale_init_value is not None:
                self.layer_scale = LayerScale2d(dim, layer_scale_init_value)
            else:
                self.layer_scale = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Reparameterized forward pass
        if self.reparam_conv is not None:
            return self.reparam_conv(x)

        # Multi-branched train-time forward pass
        return x + self.layer_scale(self.mixer(x) - self.norm(x))

    def reparameterize(self) -> None:
        if self.reparameterized is True:
            return

        self.mixer.reparameterize()
        self.norm.reparameterize()

        if isinstance(self.layer_scale, LayerScale2d):
            w = self.mixer.id_tensor + self.layer_scale.gamma.unsqueeze(-1) * (
                self.mixer.reparam_conv.weight - self.norm.reparam_conv.weight
            )
            b = torch.squeeze(self.layer_scale.gamma) * (self.mixer.reparam_conv.bias - self.norm.reparam_conv.bias)
        else:
            w = (  # type: ignore[unreachable]
                self.mixer.id_tensor + self.mixer.reparam_conv.weight - self.norm.reparam_conv.weight
            )
            b = self.mixer.reparam_conv.bias - self.norm.reparam_conv.bias

        self.reparam_conv = nn.Conv2d(
            self.dim,
            self.dim,
            kernel_size=self.kernel_size,
            stride=(1, 1),
            padding=self.kernel_size // 2,
            groups=self.dim,
            bias=True,
        )
        self.reparam_conv.weight.data = w
        self.reparam_conv.bias.data = b

        # Delete un-used branches
        for param in self.parameters():
            param.detach_()

        del self.mixer
        del self.norm
        del self.layer_scale

        self.reparameterized = True


class ConvMLP(nn.Module):
    def __init__(
        self,
        in_channels: int,
        hidden_channels: int,
        drop: float,
    ) -> None:
        super().__init__()
        self.conv = nn.Sequential()
        self.conv.add_module(
            "conv",
            nn.Conv2d(
                in_channels,
                in_channels,
                kernel_size=(7, 7),
                stride=(1, 1),
                padding=(3, 3),
                groups=in_channels,
                bias=False,
            ),
        )
        self.conv.add_module("bn", nn.BatchNorm2d(num_features=in_channels))
        self.fc1 = nn.Conv2d(in_channels, hidden_channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.act = nn.GELU()
        self.fc2 = nn.Conv2d(hidden_channels, in_channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.drop = nn.Dropout(drop)

        # Weights initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv(x)
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class RepCPE(nn.Module):
    """
    Implementation of conditional positional encoding

    For more details refer to paper:
    Conditional Positional Encodings for Vision Transformers https://arxiv.org/pdf/2102.10882.pdf

    In this implementation, we can reparameterize this module to eliminate a skip connection.
    """

    def __init__(
        self,
        in_channels: int,
        embed_dim: int,
        spatial_shape: tuple[int, int],
        reparameterized: bool,
    ) -> None:
        super().__init__()
        self.spatial_shape = spatial_shape
        self.embed_dim = embed_dim
        self.in_channels = in_channels
        self.groups = embed_dim
        self.reparameterized = reparameterized

        if reparameterized is True:
            self.reparam_conv = nn.Conv2d(
                self.in_channels,
                self.embed_dim,
                kernel_size=self.spatial_shape,
                stride=(1, 1),
                padding=self.spatial_shape[0] // 2,
                groups=self.groups,
                bias=True,
            )
        else:
            self.reparam_conv = None

            self.pe = nn.Conv2d(
                in_channels,
                embed_dim,
                kernel_size=self.spatial_shape,
                stride=(1, 1),
                padding=self.spatial_shape[0] // 2,
                groups=self.groups,
                bias=True,
            )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Reparameterized forward pass
        if self.reparam_conv is not None:
            return self.reparam_conv(x)

        # Skip connection train-time forward pass
        return self.pe(x) + x

    def reparameterize(self) -> None:
        if self.reparameterized is True:
            return

        # Build equivalent Id tensor
        input_dim = self.in_channels // self.groups
        kernel_value = torch.zeros(
            (
                self.in_channels,
                input_dim,
                self.spatial_shape[0],
                self.spatial_shape[1],
            ),
            dtype=self.pe.weight.dtype,
            device=self.pe.weight.device,
        )
        for i in range(self.in_channels):
            kernel_value[
                i,
                i % input_dim,
                self.spatial_shape[0] // 2,
                self.spatial_shape[1] // 2,
            ] = 1

        id_tensor = kernel_value

        # Reparameterize Id tensor and conv
        w_final = id_tensor + self.pe.weight
        b_final = self.pe.bias

        # Introduce reparam conv
        self.reparam_conv = nn.Conv2d(
            self.in_channels,
            self.embed_dim,
            kernel_size=self.spatial_shape,
            stride=(1, 1),
            padding=self.spatial_shape[0] // 2,
            groups=self.embed_dim,
            bias=True,
        )
        self.reparam_conv.weight.data = w_final
        self.reparam_conv.bias.data = b_final

        # Delete un-used branches
        for param in self.parameters():
            param.detach_()

        del self.pe
        self.reparameterized = True


class RepMixerBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        kernel_size: int,
        mlp_ratio: float,
        proj_drop: float,
        drop_path: float,
        layer_scale_init_value: Optional[float],
        reparameterized: bool,
    ) -> None:
        super().__init__()

        self.token_mixer = RepMixer(
            dim,
            kernel_size=kernel_size,
            layer_scale_init_value=layer_scale_init_value,
            reparameterized=reparameterized,
        )
        self.mlp = ConvMLP(
            in_channels=dim,
            hidden_channels=int(dim * mlp_ratio),
            drop=proj_drop,
        )

        if layer_scale_init_value is not None:
            self.layer_scale = LayerScale2d(dim, layer_scale_init_value)
        else:
            self.layer_scale = nn.Identity()

        self.drop_path = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.token_mixer(x)
        x = x + self.drop_path(self.layer_scale(self.mlp(x)))

        return x


class AttentionBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        mlp_ratio: float,
        proj_drop: float,
        drop_path: float,
        layer_scale_init_value: Optional[float],
    ):
        super().__init__()

        self.norm = nn.BatchNorm2d(dim)
        self.token_mixer = MHSA(dim, head_dim=32, qkv_bias=False, attn_drop=0.0, proj_drop=0.0)
        if layer_scale_init_value is not None:
            self.layer_scale_1 = LayerScale2d(dim, layer_scale_init_value)
        else:
            self.layer_scale_1 = nn.Identity()

        self.mlp = ConvMLP(
            in_channels=dim,
            hidden_channels=int(dim * mlp_ratio),
            drop=proj_drop,
        )
        if layer_scale_init_value is not None:
            self.layer_scale_2 = LayerScale2d(dim, layer_scale_init_value)
        else:
            self.layer_scale_2 = nn.Identity()

        self.drop_path = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.drop_path(self.layer_scale_1(self.token_mixer(self.norm(x))))
        x = x + self.drop_path(self.layer_scale_2(self.mlp(x)))
        return x


class FastVitStage(nn.Module):
    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(
        self,
        dim: int,
        dim_out: int,
        depth: int,
        token_mixer_type: Literal["repmixer", "attention"],
        downsample: bool,
        se_downsample: bool,
        down_patch_size: int,
        down_stride: int,
        use_cpe: bool,
        kernel_size: int,
        mlp_ratio: float,
        proj_drop_rate: float,
        drop_path_rate: list[float],
        layer_scale_init_value: Optional[float],
        reparameterized: bool,
    ) -> None:
        super().__init__()
        if downsample is True:
            self.downsample = PatchEmbed(
                patch_size=down_patch_size,
                stride=down_stride,
                in_channels=dim,
                embed_dim=dim_out,
                use_se=se_downsample,
                reparameterized=reparameterized,
            )
        else:
            assert dim == dim_out
            self.downsample = nn.Identity()

        if use_cpe is True:
            self.pos_emb = RepCPE(dim_out, dim_out, spatial_shape=(7, 7), reparameterized=reparameterized)
        else:
            self.pos_emb = nn.Identity()

        blocks = []
        for block_idx in range(depth):
            if token_mixer_type == "repmixer":  # nosec
                blocks.append(
                    RepMixerBlock(
                        dim_out,
                        kernel_size=kernel_size,
                        mlp_ratio=mlp_ratio,
                        proj_drop=proj_drop_rate,
                        drop_path=drop_path_rate[block_idx],
                        layer_scale_init_value=layer_scale_init_value,
                        reparameterized=reparameterized,
                    )
                )
            elif token_mixer_type == "attention":  # nosec
                blocks.append(
                    AttentionBlock(
                        dim_out,
                        mlp_ratio=mlp_ratio,
                        proj_drop=proj_drop_rate,
                        drop_path=drop_path_rate[block_idx],
                        layer_scale_init_value=layer_scale_init_value,
                    )
                )
            else:
                raise ValueError(f"Token mixer type: {token_mixer_type} not supported")

        self.blocks = nn.Sequential(*blocks)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.downsample(x)
        x = self.pos_emb(x)
        x = self.blocks(x)

        return x


class FastViT(DetectorBackbone, PreTrainEncoder, MaskedTokenRetentionMixin):
    default_size = (256, 256)
    block_group_regex = r"body\.stage(\d+)\.blocks\.(\d+)"

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        self.reparameterized = False
        cls_ratio = 2.0
        layers: tuple[int, int, int, int] = self.config["layers"]
        embed_dims: tuple[int, int, int, int] = self.config["embed_dims"]
        mlp_ratios: tuple[int, int, int, int] = self.config["mlp_ratios"]
        se_downsamples: tuple[bool, bool, bool, bool] = self.config["se_downsamples"]
        use_cpe: tuple[bool, bool, bool, bool] = self.config["use_cpe"]
        token_mixers: tuple[str, str, str, str] = self.config["token_mixers"]
        layer_scale_init_value: float = self.config["layer_scale_init_value"]
        drop_path_rate: float = self.config["drop_path_rate"]

        self.stem = nn.Sequential(
            MobileOneBlock(
                in_channels=self.input_channels,
                out_channels=embed_dims[0],
                kernel_size=3,
                stride=2,
                padding=1,
                groups=1,
                use_se=False,
                use_act=True,
                use_scale_branch=True,
                num_conv_branches=1,
                reparameterized=self.reparameterized,
                activation_layer=nn.GELU,
            ),
            MobileOneBlock(
                in_channels=embed_dims[0],
                out_channels=embed_dims[0],
                kernel_size=3,
                stride=2,
                padding=1,
                groups=embed_dims[0],
                use_se=False,
                use_act=True,
                use_scale_branch=True,
                num_conv_branches=1,
                reparameterized=self.reparameterized,
                activation_layer=nn.GELU,
            ),
            MobileOneBlock(
                in_channels=embed_dims[0],
                out_channels=embed_dims[0],
                kernel_size=1,
                stride=1,
                padding=0,
                groups=1,
                use_se=False,
                use_act=True,
                use_scale_branch=True,
                num_conv_branches=1,
                reparameterized=self.reparameterized,
                activation_layer=nn.GELU,
            ),
        )

        num_stages = len(layers)
        prev_dim = embed_dims[0]
        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(layers)).split(layers)]
        downsamples = (False,) + (True,) * (num_stages - 1)
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i in range(num_stages):
            downsample = downsamples[i] or prev_dim != embed_dims[i]
            stages[f"stage{i+1}"] = FastVitStage(
                dim=prev_dim,
                dim_out=embed_dims[i],
                depth=layers[i],
                token_mixer_type=token_mixers[i],  # type: ignore[arg-type]
                downsample=downsample,
                se_downsample=se_downsamples[i],
                down_patch_size=7,
                down_stride=2,
                use_cpe=use_cpe[i],
                kernel_size=3,
                mlp_ratio=mlp_ratios[i],
                proj_drop_rate=0.0,
                drop_path_rate=dpr[i],
                layer_scale_init_value=layer_scale_init_value,
                reparameterized=self.reparameterized,
            )
            return_channels.append(embed_dims[i])
            prev_dim = embed_dims[i]

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            MobileOneBlock(
                in_channels=embed_dims[-1],
                out_channels=int(embed_dims[-1] * cls_ratio),
                kernel_size=3,
                stride=1,
                padding=1,
                groups=embed_dims[-1],
                use_se=True,
                use_act=True,
                use_scale_branch=True,
                num_conv_branches=1,
                reparameterized=self.reparameterized,
                activation_layer=nn.GELU,
            ),
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = int(embed_dims[-1] * cls_ratio)
        self.classifier = self.create_classifier()

        self.stem_stride = 4
        self.stem_width = embed_dims[0]
        self.encoding_size = int(embed_dims[-1] * cls_ratio)

        # Weights initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        x = self.stem(x)
        x = mask_tensor(x, mask, patch_factor=self.max_stride // self.stem_stride, mask_token=mask_token)
        x = self.body(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            result["features"] = x
        if return_keys in ("all", "embedding"):
            result["embedding"] = self.features(x)

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def reparameterize_model(self) -> None:
        for module in self.modules():
            if hasattr(module, "reparameterize") is True:
                module.reparameterize()

        self.reparameterized = True


registry.register_model_config(
    "fastvit_t8",
    FastViT,
    config={
        "layers": (2, 2, 4, 2),
        "embed_dims": (48, 96, 192, 384),
        "mlp_ratios": (3, 3, 3, 3),
        "se_downsamples": (False, False, False, False),
        "use_cpe": (False, False, False, False),
        "token_mixers": ("repmixer", "repmixer", "repmixer", "repmixer"),
        "layer_scale_init_value": 1e-5,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "fastvit_t12",
    FastViT,
    config={
        "layers": (2, 2, 6, 2),
        "embed_dims": (64, 128, 256, 512),
        "mlp_ratios": (3, 3, 3, 3),
        "se_downsamples": (False, False, False, False),
        "use_cpe": (False, False, False, False),
        "token_mixers": ("repmixer", "repmixer", "repmixer", "repmixer"),
        "layer_scale_init_value": 1e-5,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "fastvit_s12",
    FastViT,
    config={
        "layers": (2, 2, 6, 2),
        "embed_dims": (64, 128, 256, 512),
        "mlp_ratios": (4, 4, 4, 4),
        "se_downsamples": (False, False, False, False),
        "use_cpe": (False, False, False, False),
        "token_mixers": ("repmixer", "repmixer", "repmixer", "repmixer"),
        "layer_scale_init_value": 1e-5,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "fastvit_sa12",
    FastViT,
    config={
        "layers": (2, 2, 6, 2),
        "embed_dims": (64, 128, 256, 512),
        "mlp_ratios": (4, 4, 4, 4),
        "se_downsamples": (False, False, False, False),
        "use_cpe": (False, False, False, True),
        "token_mixers": ("repmixer", "repmixer", "repmixer", "attention"),
        "layer_scale_init_value": 1e-5,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "fastvit_sa24",
    FastViT,
    config={
        "layers": (4, 4, 12, 4),
        "embed_dims": (64, 128, 256, 512),
        "mlp_ratios": (4, 4, 4, 4),
        "se_downsamples": (False, False, False, False),
        "use_cpe": (False, False, False, True),
        "token_mixers": ("repmixer", "repmixer", "repmixer", "attention"),
        "layer_scale_init_value": 1e-5,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "fastvit_sa36",
    FastViT,
    config={
        "layers": (6, 6, 18, 6),
        "embed_dims": (64, 128, 256, 512),
        "mlp_ratios": (4, 4, 4, 4),
        "se_downsamples": (False, False, False, False),
        "use_cpe": (False, False, False, True),
        "token_mixers": ("repmixer", "repmixer", "repmixer", "attention"),
        "layer_scale_init_value": 1e-6,
        "drop_path_rate": 0.2,
    },
)
registry.register_model_config(
    "fastvit_ma36",
    FastViT,
    config={
        "layers": (6, 6, 18, 6),
        "embed_dims": (76, 152, 304, 608),
        "mlp_ratios": (4, 4, 4, 4),
        "se_downsamples": (False, False, False, False),
        "use_cpe": (False, False, False, True),
        "token_mixers": ("repmixer", "repmixer", "repmixer", "attention"),
        "layer_scale_init_value": 1e-6,
        "drop_path_rate": 0.35,
    },
)

registry.register_model_config(
    "mobileclip_i0",
    FastViT,
    config={
        "layers": (2, 6, 10, 2),
        "embed_dims": (64, 128, 256, 512),
        "mlp_ratios": (3, 3, 3, 3),
        "se_downsamples": (False, False, True, True),
        "use_cpe": (False, False, False, True),
        "token_mixers": ("repmixer", "repmixer", "repmixer", "attention"),
        "layer_scale_init_value": 1e-5,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "mobileclip_i1",
    FastViT,
    config={
        "layers": (4, 12, 20, 4),
        "embed_dims": (64, 128, 256, 512),
        "mlp_ratios": (3, 3, 3, 3),
        "se_downsamples": (False, False, True, True),
        "use_cpe": (False, False, False, True),
        "token_mixers": ("repmixer", "repmixer", "repmixer", "attention"),
        "layer_scale_init_value": 1e-5,
        "drop_path_rate": 0.05,
    },
)
registry.register_model_config(
    "mobileclip_i2",
    FastViT,
    config={
        "layers": (4, 12, 24, 4),
        "embed_dims": (80, 160, 320, 640),
        "mlp_ratios": (3, 3, 3, 3),
        "se_downsamples": (False, False, True, True),
        "use_cpe": (False, False, False, True),
        "token_mixers": ("repmixer", "repmixer", "repmixer", "attention"),
        "layer_scale_init_value": 1e-5,
        "drop_path_rate": 0.15,
    },
)

registry.register_weights(
    "fastvit_t8_il-common",
    {
        "description": "FastViT T-8 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 13.8,
                "sha256": "ec976d4edebf2215807de264e67d8bf65aa6da893295d4132cd418ea42b2171f",
            }
        },
        "net": {"network": "fastvit_t8", "tag": "il-common"},
    },
)
registry.register_weights(
    "fastvit_t8_il-common_reparameterized",
    {
        "description": "FastViT T-8 (reparameterized) model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 13.5,
                "sha256": "062c98f4acd50f726346a572a6f505c1b804c8ca6618da8c3719ff6764e9ac3b",
            }
        },
        "net": {"network": "fastvit_t8", "tag": "il-common_reparameterized", "reparameterized": True},
    },
)
registry.register_weights(
    "fastvit_t12_il-common",
    {
        "description": "FastViT T-12 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 26.7,
                "sha256": "66a61c8ba504357869af29349b3554b28567f020b539ad7c89d7b4d90de9b490",
            }
        },
        "net": {"network": "fastvit_t12", "tag": "il-common"},
    },
)
registry.register_weights(
    "fastvit_t12_il-common_reparameterized",
    {
        "description": "FastViT T-12 (reparameterized) model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 26.3,
                "sha256": "debb0dc5a96c9bb843e57bdea995c0b1535aceb0ad26aa8fb5d23ecdb8427f9b",
            }
        },
        "net": {"network": "fastvit_t12", "tag": "il-common_reparameterized", "reparameterized": True},
    },
)
registry.register_weights(
    "fastvit_sa12_il-common",
    {
        "description": "FastViT SA-12 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 42.0,
                "sha256": "5478786787fe8f23cff31216de941e6d4b0aab307ea0b9051115f7691906707f",
            }
        },
        "net": {"network": "fastvit_sa12", "tag": "il-common"},
    },
)
registry.register_weights(
    "fastvit_sa12_il-common_reparameterized",
    {
        "description": "FastViT SA-12 (reparameterized) model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 41.7,
                "sha256": "6e83955ac63b323760be62383c21ef72417c7bca461347401c9cd0d1d424ff62",
            }
        },
        "net": {"network": "fastvit_sa12", "tag": "il-common_reparameterized", "reparameterized": True},
    },
)
registry.register_weights(
    "mobileclip_i0_il-common",
    {
        "description": "MobileClip i0 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 43.5,
                "sha256": "f6599c5c5373d0928add5747c9e6524d5273b24c3c580a1c7b80acc7c10fd655",
            }
        },
        "net": {"network": "mobileclip_i0", "tag": "il-common"},
    },
)
registry.register_weights(
    "mobileclip_i0_il-common_reparameterized",
    {
        "description": "MobileClip i0 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 42.9,
                "sha256": "aec9bf62c0eaa12943158b056e7cef2fa93ebd78519080ad18bfacc5ef006670",
            }
        },
        "net": {"network": "mobileclip_i0", "tag": "il-common_reparameterized", "reparameterized": True},
    },
)



import argparse
import fnmatch
import logging
from pathlib import Path
from typing import Any

import numpy as np
import polars as pl
import polars.datatypes.classes
from rich.console import Console
from rich.table import Table

from birder.common import cli
from birder.conf import settings
from birder.results.classification import Results
from birder.results.classification import SparseResults
from birder.results.classification import compare_results
from birder.results.classification import detect_file_format
from birder.results.classification import load_results
from birder.results.gui import ROC
from birder.results.gui import ConfusionMatrix
from birder.results.gui import PrecisionRecall
from birder.results.gui import ProbabilityHistogram

logger = logging.getLogger(__name__)


def print_per_class_report(results_dict: dict[str, Results], classes: list[str]) -> None:
    console = Console()

    # Expand classes according to shell-style wildcards
    all_classes = []
    for results in results_dict.values():
        for cls in classes:
            all_classes.extend(fnmatch.filter(results.label_names, cls))

    classes = sorted(list(set(all_classes)))

    # Per class
    table = Table(show_header=True, header_style="bold dark_magenta")
    table.add_column("File name")
    table.add_column("Class name", style="dim")
    table.add_column("Precision", justify="right")
    table.add_column("Recall", justify="right")
    table.add_column("F1-score", justify="right")
    table.add_column("Samples", justify="right")
    table.add_column("False negative", justify="right")
    table.add_column("False positive", justify="right")

    for cls in classes:
        for name, results in results_dict.items():
            report_df = results.detailed_report()
            row = report_df.filter(pl.col("Class name") == cls)
            if row.is_empty() is True:
                continue

            recall_msg = f"{row['Recall'][0]:.4f}"
            if row["Recall"][0] < 0.75:
                recall_msg = "[red1]" + recall_msg + "[/red1]"
            elif row["Recall"][0] < 0.9:
                recall_msg = "[dark_orange]" + recall_msg + "[/dark_orange]"

            f1_msg = f"{row['F1-score'][0]:.4f}"
            if row["F1-score"][0] == 1.0:
                f1_msg = "[green]" + f1_msg + "[/green]"

            table.add_row(
                name,
                row["Class name"][0],
                f"{row['Precision'][0]:.4f}",
                recall_msg,
                f1_msg,
                f"{row['Samples'][0]}",
                f"{row['False negative'][0]}",
                f"{row['False positive'][0]}",
            )

    console.print(table)


def print_report(results_dict: dict[str, Results]) -> None:
    if len(results_dict) == 1:
        results = next(iter(results_dict.values()))
        results.pretty_print()
        return

    results_df = compare_results(results_dict)
    console = Console()
    table = Table(show_header=True, header_style="bold dark_magenta")
    for idx, column in enumerate(results_df.columns):
        if idx == 0:
            table.add_column(column)
        else:
            table.add_column(column, justify="right")

        if isinstance(results_df[column].dtype, polars.datatypes.classes.FloatType):
            results_df = results_df.with_columns(
                pl.col(column).map_elements(lambda x: f"{x:.4f}", return_dtype=pl.String)
            )
        else:
            results_df = results_df.with_columns(pl.col(column).cast(pl.String))

    for row in results_df.iter_rows():
        table.add_row(*row)

    console.print(table)
    console.print("\n")


def print_most_confused_pairs(results: Results) -> None:
    most_confused_df = results.most_confused_pairs(n=14)

    console = Console()

    table = Table(show_header=True, header_style="bold dark_magenta")
    for column in most_confused_df.columns:
        if isinstance(most_confused_df[column].dtype, polars.datatypes.classes.NumericType):
            table.add_column(column.capitalize(), justify="right")
        else:
            table.add_column(column.capitalize())

    for row in most_confused_df.iter_rows():
        table.add_row(row[0], row[1], str(row[2]), str(row[3]))

    console.print(table)


def convert_to_sparse(results_file: str, sparse_k: int) -> None:
    logger.info(f"Converting {results_file} to sparse format (k={sparse_k})...")
    (_, detected_sparse_k) = detect_file_format(results_file)

    if detected_sparse_k is not None:
        logger.info(f"File is already in sparse format (with k={detected_sparse_k}). Skipping conversion.")
        return

    sparse_results = SparseResults.load(results_file, sparse_k=sparse_k)

    input_path = Path(results_file)
    if input_path.suffix == ".gz":
        input_path = input_path.with_suffix("")

    output_path = input_path.with_stem(f"{input_path.stem}_sparse")
    if output_path.exists() is True:
        logger.warning(f"Target file already exists: {output_path}. Skipping conversion.")
        return

    relative_path = output_path.relative_to(settings.RESULTS_DIR)
    sparse_results.save(str(relative_path))


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "results",
        allow_abbrev=False,
        help="read and process result files",
        description="read and process result files",
        epilog=(
            "Usage examples:\n"
            "python -m birder.tools results results/vit_l16_mim340_218_e0_448px_crop1.0_10883.csv "
            "--cnf --cnf-errors-only\n"
            'python -m birder.tools results results/deit_2_* --print --classes "Lesser kestrel" '
            '"Common kestrel" "*swan"\n'
            "python -m birder.tools results results/inception_resnet_v2_105_e100_299px_crop1.0_3150.csv "
            "--print --roc\n"
            "python -m birder.tools results results/inception_resnet_v2_105_e100_299px_crop1.0_3150.csv "
            '--pr-curve --classes "Common crane" "Demoiselle crane"\n'
            "python -m birder.tools results results/densenet_121_105_e100_224px_crop1.0_3150.csv --prob-hist "
            '"Common kestrel" "Red-footed falcon"\n'
            "python -m birder.tools results results/inception_resnet_v2_105_e100_299px_crop1.0_3150.csv --cnf "
            "--classes Mallard Unknown Wallcreeper\n"
            "python -m birder.tools results results/maxvit_2_154_e0_288px_crop1.0_6286.csv "
            "results/inception_next_1_160_e0_384px_crop1.0_6762.csv --print\n"
            "python -m birder.tools results results/convnext_v2_base_214_e0_448px_crop1.0_10682.csv "
            '--prob-hist "Common kestrel" "Lesser kestrel"\n'
            "python -m birder.tools results results/squeezenet_il-common_367_e0_259px_crop1.0_13029.csv "
            "--most-confused\n"
            "python -m birder.tools results results/inat21/rope_vit_b14_inat21_10000_336px_crop1.0_100000.csv.gz "
            "--to-sparse --sparse-k 15\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument("--print", default=False, action="store_true", help="print results table")
    subparser.add_argument("--short-print", default=False, action="store_true", help="print results")
    subparser.add_argument("--save-summary", default=False, action="store_true", help="save results summary as csv")
    subparser.add_argument("--summary-suffix", type=str, help="add suffix to summary file")
    subparser.add_argument(
        "--imperfect-only",
        default=False,
        action="store_true",
        help="display only classes with imperfect performance (F1 < 1.0)",
    )
    subparser.add_argument("--classes", default=[], type=str, nargs="+", help="class names to compare")
    subparser.add_argument("--list-mistakes", default=False, action="store_true", help="list all mistakes")
    subparser.add_argument("--list-out-of-k", default=False, action="store_true", help="list all samples not in top-k")
    subparser.add_argument("--cnf", default=False, action="store_true", help="plot confusion matrix")
    subparser.add_argument(
        "--cnf-errors-only",
        default=False,
        action="store_true",
        help="show only classes with mistakes at the confusion matrix",
    )
    subparser.add_argument("--cnf-save", default=False, action="store_true", help="save confusion matrix as csv")
    subparser.add_argument("--roc", default=False, action="store_true", help="plot roc curve")
    subparser.add_argument("--pr-curve", default=False, action="store_true", help="plot precision recall curve")
    subparser.add_argument(
        "--prob-hist", type=str, nargs=2, help="classes to plot probability histogram against each other"
    )
    subparser.add_argument("--most-confused", default=False, action="store_true", help="print most confused pairs")
    subparser.add_argument(
        "--to-sparse", default=False, action="store_true", help="convert regular results file to sparse format"
    )
    subparser.add_argument("--sparse-k", type=int, default=10, help="number of top probabilities to keep per sample")
    subparser.add_argument("result_files", type=str, nargs="+", help="result files to process")
    subparser.set_defaults(func=main)


# pylint: disable=too-many-branches
def main(args: argparse.Namespace) -> None:
    if args.to_sparse is True:
        # Type conversion exits early, no further flag handling is done
        logger.info(f"Converting {len(args.result_files)} file(s) to sparse format...")
        for results_file in args.result_files:
            convert_to_sparse(results_file, args.sparse_k)

        return

    results_dict: dict[str, Results | SparseResults] = {}
    for results_file in args.result_files:
        results = load_results(results_file)
        result_name = results_file.split("/")[-1]
        results_dict[result_name] = results

    if args.print is True:
        if args.imperfect_only is True and len(results_dict) > 1:
            logger.warning("Cannot print mistakes in compare mode. processing only the first file")

        if args.imperfect_only is True:
            (result_name, results) = next(iter(results_dict.items()))
            mistake_prediction_indices = results.mistakes["prediction"].unique().to_numpy().tolist()
            mistake_label_indices = results.mistakes["label"].unique().to_numpy().tolist()
            imperfect_class_indices = np.unique(mistake_prediction_indices + mistake_label_indices).tolist()

            new_results = results.filter_by_labels(imperfect_class_indices)
            results_dict = {result_name: new_results}

        print_report(results_dict)
        if len(args.classes) > 0:
            print_per_class_report(results_dict, args.classes)

    if args.short_print is True:
        for name, results in results_dict.items():
            print(f"{name}: {results}\n")

    if args.save_summary is True:
        if args.summary_suffix is not None:
            summary_path = settings.RESULTS_DIR.joinpath(f"summary_{args.summary_suffix}.csv")
        else:
            summary_path = settings.RESULTS_DIR.joinpath("summary.csv")

        if summary_path.exists() is True:
            logger.warning(f"Summary already exists '{summary_path}', skipping...")
        else:
            logger.info(f"Writing results summary at '{summary_path}...")
            results_df = compare_results(results_dict)
            results_df.write_csv(summary_path)

    if args.list_mistakes is True:
        for name, results in results_dict.items():
            mistakes = sorted(list(results.mistakes["sample"]))
            print("\n".join(mistakes))
            logger.info(f"{len(results.mistakes):,} mistakes found at {name}")

    if args.list_out_of_k is True:
        for name, results in results_dict.items():
            out_of_k = sorted(list(results.out_of_top_k["sample"]))
            print("\n".join(out_of_k))
            logger.info(f"{len(results.out_of_top_k):,} out of k found at {name}")

    if args.cnf is True:
        if len(results_dict) > 1:
            logger.warning("Cannot compare confusion matrix, processing only the first file")

        results = next(iter(results_dict.values()))
        if len(args.classes) > 0:
            selected_class_indices = []
            for pattern in args.classes:
                selected_class_indices.extend(
                    [idx for idx, name in enumerate(results.label_names) if fnmatch.fnmatch(name, pattern)]
                )

            cnf_results = results.filter_by_labels(np.unique(selected_class_indices).tolist())

        elif args.cnf_errors_only is True:
            mistake_prediction_indices = results.mistakes["prediction"].unique().to_numpy().tolist()
            mistake_label_indices = results.mistakes["label"].unique().to_numpy().tolist()
            imperfect_class_indices = np.unique(mistake_prediction_indices + mistake_label_indices).tolist()

            cnf_results = results.filter_by_labels(imperfect_class_indices)

        else:
            cnf_results = results

        ConfusionMatrix(cnf_results).show()

    if args.cnf_save is True:
        for results_file, results in results_dict.items():
            filename = f"{results_file[:-4]}_confusion_matrix.csv"
            ConfusionMatrix(results).save(filename)

    if args.roc is True:
        roc = ROC()
        for name, results in results_dict.items():
            if isinstance(results, SparseResults):
                logger.warning(
                    f"Skipping ROC curve for '{name}' as it is a SparseResults object. "
                    "ROC curve requires full probability outputs."
                )
                continue

            roc.add_result(Path(name).name, results)

        roc.show(args.classes)

    if args.pr_curve is True:
        pr_curve = PrecisionRecall()
        for name, results in results_dict.items():
            if isinstance(results, SparseResults):
                logger.warning(
                    f"Skipping Precision-Recall curve for '{name}' as it is a SparseResults object. "
                    "Precision-Recall curve requires full probability outputs."
                )
                continue
            pr_curve.add_result(Path(name).name, results)

        pr_curve.show(args.classes)

    if args.prob_hist is not None:
        if len(results_dict) > 1:
            logger.warning("Cannot compare probability histograms, processing only the first file")

        results = next(iter(results_dict.values()))
        if isinstance(results, SparseResults):
            logger.warning(
                "Probability histogram is not supported for SparseResults objects. "
                "It requires full probability outputs."
            )
            return  # Exit early as this feature is not supported for sparse data

        ProbabilityHistogram(results).show(*args.prob_hist)

    if args.most_confused is True:
        if len(results_dict) > 1:
            logger.warning("Cannot compare, processing only the first file")

        results = next(iter(results_dict.values()))
        print_most_confused_pairs(results)



import logging
import math
import unittest

import torch

from birder import net
from birder import optim

logging.disable(logging.CRITICAL)


class TestOptimizer(unittest.TestCase):
    def test_lamb(self) -> None:
        # Just a simple sanity test
        size = (64, 64)
        n = net.MobileNet_v1(3, 10, config={"alpha": 0.5}, size=size)
        optimizer = optim.Lamb(n.parameters(), lr=0.1)
        repr(optimizer)

        for _ in range(5):
            optimizer.zero_grad()
            out = n.embedding(torch.rand((4, 3, *size)))
            loss = out.abs().sum()
            loss.backward()
            optimizer.step()

            self.assertFalse(math.isnan(loss.item()))

    def test_lars(self) -> None:
        # Just a simple sanity test
        size = (64, 64)
        n = net.MobileNet_v1(3, 10, config={"alpha": 0.5}, size=size)
        optimizer = optim.Lars(n.parameters(), lr=0.1)
        repr(optimizer)

        for _ in range(5):
            optimizer.zero_grad()
            out = n.embedding(torch.rand((4, 3, *size)))
            loss = out.abs().sum()
            loss.backward()
            optimizer.step()

            self.assertFalse(math.isnan(loss.item()))



import argparse
import json
import logging
import os
from datetime import date
from pathlib import Path
from typing import Any

from tqdm import tqdm

import birder
from birder.common import cli
from birder.common import fs_ops
from birder.common import lib
from birder.conf import settings

logger = logging.getLogger(__name__)


def _create_annotation(
    points: tuple[tuple[float, float], tuple[float, float]], label: int, image_id: int, annotation_id: int
) -> dict[str, Any]:
    annotation: dict[str, Any] = {}
    annotation["iscrowd"] = 0
    annotation["image_id"] = image_id

    # Bounding box in (x, y, w, h) format
    (x0, y0) = points[0]
    (x1, y1) = points[1]
    x = min(x0, x1)
    y = min(y0, y1)
    w = abs(x0 - x1)
    h = abs(y0 - y1)
    annotation["bbox"] = [x, y, w, h]
    annotation["category_id"] = label
    annotation["id"] = annotation_id

    return annotation


def labelme_to_coco(args: argparse.Namespace) -> None:
    prefix = str(settings.DETECTION_DATA_PATH) + "/"
    if args.class_file is not None:
        class_file = args.class_file
        base_name = Path(args.data_path).stem + "_" + Path(args.class_file).stem
    else:
        class_file = settings.DETECTION_DATA_PATH.joinpath(settings.CLASS_LIST_NAME)
        base_name = Path(args.data_path).stem

    target_path = Path(args.data_path).parent.joinpath(f"{base_name}_coco.json")

    class_to_idx = fs_ops.read_class_file(class_file)
    class_to_idx = lib.detection_class_to_idx(class_to_idx)

    image_list = []
    annotation_list = []
    annotation_id = 0
    for idx, json_path in tqdm(enumerate(fs_ops.file_iter(args.data_path, extensions=[".json"])), leave=False):
        with open(json_path, "r", encoding="utf-8") as handle:
            data = json.load(handle)

        unknown = data["flags"].get("unknown", False)
        if unknown is True and args.include_unknown is False:
            continue

        if args.class_file is not None:
            skip = False
            for shapes in data["shapes"]:
                if shapes["label"] not in class_to_idx:
                    logger.debug(f"Found unknown label: {shapes['label']}, skipping image {data['imagePath']}")
                    skip = True
                    break

            if skip is True:
                continue

        image_path = Path(json_path).parent.joinpath(data["imagePath"])
        image = {
            "id": idx,
            "width": data["imageWidth"],
            "height": data["imageHeight"],
            "file_name": os.path.normpath(image_path).removeprefix(prefix),
        }
        image_list.append(image)

        for shapes in data["shapes"]:
            if shapes["shape_type"] != "rectangle":
                logger.error("Only detection rectangles are supported, aborting...")
                raise SystemExit(1)

            label = shapes["label"]
            if label not in class_to_idx:
                logger.error(f"Found unknown label: {label}, aborting...")
                raise SystemExit(1)

            points = shapes["points"]
            annotation_list.append(_create_annotation(points, class_to_idx[label], idx, annotation_id))
            annotation_id += 1

    # Create categories
    category_list = []
    for class_name, class_id in class_to_idx.items():
        category_list.append(
            {
                "supercategory": class_name,
                "id": class_id,
                "name": class_name,
            }
        )

    # Create COCO format dictionary
    coco: dict[str, Any] = {}
    coco["info"] = {
        "version": "1.0",
        "birder_version": birder.__version__,
        "year": date.today().year,
        "date_created": date.today().isoformat(),
    }
    coco["images"] = image_list
    coco["categories"] = category_list
    coco["annotations"] = annotation_list

    # Save
    logger.info(f"Saving COCO file at {target_path}...")
    with open(target_path, "w", encoding="utf-8") as handle:
        json.dump(coco, handle, indent=2)

    logger.info(f"Written {len(image_list)} images with {len(annotation_list)} annotations")


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "labelme-to-coco",
        allow_abbrev=False,
        help="convert labelme detection annotations to coco format",
        description="convert labelme detection annotations to coco format",
        epilog=(
            "Usage examples:\n"
            "python -m birder.tools labelme-to-coco data/detection_data/training_annotations\n"
            "python -m birder.tools labelme-to-coco --class-file data/il-common_classes.txt "
            "data/detection_data/training_annotations\n"
            "python -m birder.tools labelme-to-coco --include-unknown data/detection_data/validation_annotations\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument(
        "--include-unknown", default=False, action="store_true", help="include files with unknown flag"
    )
    subparser.add_argument("--class-file", type=str, help="class list file")
    subparser.add_argument("data_path", help="image directory path")
    subparser.set_defaults(func=main)


def main(args: argparse.Namespace) -> None:
    labelme_to_coco(args)



from collections.abc import Callable

import torch
from torch import nn
from torchvision.ops import MLP


class FFN(MLP):
    """
    Just a simple adaptor for interface consistency
    """

    def __init__(
        self,
        in_features: int,
        hidden_features: int,
        act_layer: Callable[..., nn.Module] = nn.ReLU,
        bias: bool = True,
        dropout: float = 0.0,
    ):
        super().__init__(
            in_features,
            [hidden_features, in_features],
            activation_layer=act_layer,
            inplace=None,
            bias=bias,
            dropout=dropout,
        )


# pylint: disable=invalid-name
class SwiGLU_FFN(nn.Module):
    """
    Paper "GLU Variants Improve Transformer", https://arxiv.org/abs/2002.05202
    Adapted from: https://github.com/huggingface/pytorch-image-models/blob/main/timm/layers/mlp.py
    """

    def __init__(
        self,
        in_features: int,
        hidden_features: int,
        act_layer: Callable[..., nn.Module] = nn.SiLU,
        bias: bool = True,
        dropout: float = 0.0,
    ):
        super().__init__()
        self.fc1_g = nn.Linear(in_features, hidden_features, bias=bias)
        self.fc1_x = nn.Linear(in_features, hidden_features, bias=bias)
        self.act = act_layer()
        self.drop1 = nn.Dropout(dropout)
        self.fc2 = nn.Linear(hidden_features, in_features, bias=bias)
        self.drop2 = nn.Dropout(dropout)

        # Weight initialization
        nn.init.normal_(self.fc1_g.weight, std=1e-6)
        if self.fc1_g.bias is not None:
            nn.init.ones_(self.fc1_g.bias)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x_gate = self.fc1_g(x)
        x = self.fc1_x(x)
        x = self.act(x_gate) * x
        x = self.drop1(x)
        x = self.fc2(x)
        x = self.drop2(x)

        return x



from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import torch
from PIL import Image
from torch import nn


def show_mask_on_image(
    img: npt.NDArray[np.float32], mask: npt.NDArray[np.float32], image_weight: float = 0.5
) -> npt.NDArray[np.float32]:
    color_map = matplotlib.colormaps["jet"]
    heatmap = color_map(mask)[:, :, :3]

    cam: npt.NDArray[np.float32] = (1 - image_weight) * heatmap + image_weight * img
    cam = cam / np.max(cam)
    cam = cam * 255

    return cam.astype(np.uint8)


@dataclass
class InterpretabilityResult:
    original_image: npt.NDArray[np.float32]
    visualization: npt.NDArray[np.float32] | npt.NDArray[np.uint8]
    raw_output: npt.NDArray[np.float32]

    def show(self, figsize: tuple[int, int] = (12, 8)) -> None:
        _, (ax1, ax2) = plt.subplots(1, 2, figsize=figsize)
        ax1.imshow(self.visualization)
        ax2.imshow(self.original_image)
        plt.show()


class Interpreter:
    def __init__(self, model: nn.Module, device: torch.device, transform: Callable[..., torch.Tensor]) -> None:
        self.model = model.eval()
        self.device = device
        self.transform = transform

    def interpret(self, image: str | Path | Image.Image, target_class: Optional[int] = None) -> InterpretabilityResult:
        raise NotImplementedError

    def _preprocess_image(self, image: str | Path | Image.Image) -> tuple[torch.Tensor, npt.NDArray[np.float32]]:
        if isinstance(image, (str, Path)):
            image = Image.open(image)

        # Transform for model
        input_tensor = self.transform(image).unsqueeze(dim=0).to(self.device)

        # Store original for visualization
        rgb_img = np.array(image.resize(input_tensor.shape[-2:])).astype(np.float32) / 255.0

        return (input_tensor, rgb_img)



"""
GroupMixFormer, adapted from
https://github.com/AILab-CVC/GroupMixFormer/blob/main/models/groupmixformer.py

Paper "Advancing Vision Transformers with Group-Mix Attention", https://arxiv.org/abs/2311.15157
"""

# Reference license: MIT

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import StochasticDepth

from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class SeparableConv2d(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        padding: tuple[int, int],
        bias: bool = False,
    ) -> None:
        super().__init__()
        self.conv1 = nn.Conv2d(
            in_channels,
            in_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            groups=in_channels,
            bias=bias,
        )
        self.pointwise_conv = nn.Conv2d(
            in_channels, out_channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=bias
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.pointwise_conv(self.conv1(x))


class Agg0(nn.Module):
    def __init__(self, seg_dim: int) -> None:
        super().__init__()
        self.conv = SeparableConv2d(seg_dim * 3, seg_dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.norm = nn.LayerNorm(seg_dim)
        self.act = nn.Hardswish()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv(x)
        (B, C, _, _) = x.size()
        x = self.act(self.norm(x.reshape(B, C, -1).permute(0, 2, 1)))

        return x


class Aggregator(nn.Module):
    def __init__(self, dim: int, seg: int) -> None:
        super().__init__()
        self.dim = dim
        self.seg = seg
        self.seg_dim = self.dim // self.seg

        self.norm0 = nn.BatchNorm2d(self.seg_dim)
        self.act0 = nn.Hardswish()

        self.agg1 = SeparableConv2d(self.seg_dim, self.seg_dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.norm1 = nn.BatchNorm2d(self.seg_dim)
        self.act1 = nn.Hardswish()

        self.agg2 = SeparableConv2d(self.seg_dim, self.seg_dim, kernel_size=(5, 5), stride=(1, 1), padding=(2, 2))
        self.norm2 = nn.BatchNorm2d(self.seg_dim)
        self.act2 = nn.Hardswish()

        self.agg3 = SeparableConv2d(self.seg_dim, self.seg_dim, kernel_size=(7, 7), stride=(1, 1), padding=(3, 3))
        self.norm3 = nn.BatchNorm2d(self.seg_dim)
        self.act3 = nn.Hardswish()

        self.agg0 = Agg0(self.seg_dim)

    def forward(self, x: torch.Tensor, size: tuple[int, int], num_head: int) -> tuple[torch.Tensor, torch.Tensor]:
        (B, _, C) = x.size()
        (H, W) = size
        # assert N == H * W

        x = x.transpose(1, 2).view(B, C, H, W)
        xs = x.split([self.seg_dim] * self.seg, dim=1)

        x_local = (
            xs[4].reshape(3, B // 3, self.seg_dim, H, W).permute(1, 0, 2, 3, 4).reshape(B // 3, 3 * self.seg_dim, H, W)
        )
        x_local = self.agg0(x_local)

        x0 = self.act0(self.norm0(xs[0]))
        x1 = self.act1(self.norm1(self.agg1(xs[1])))
        x2 = self.act2(self.norm2(self.agg2(xs[2])))
        x3 = self.act3(self.norm3(self.agg3(xs[3])))

        x = torch.concat([x0, x1, x2, x3], dim=1)

        C = C // 5 * 4
        x = x.reshape(3, B // 3, num_head, C // num_head, H * W).permute(0, 1, 2, 4, 3)

        return (x, x_local)


class ConvRelPosEnc(nn.Module):
    def __init__(self, channels: int, window: dict[int, int]) -> None:
        super().__init__()

        self.conv_list = nn.ModuleList()
        head_splits = []
        for cur_window, cur_head_split in window.items():
            padding_size = cur_window // 2
            cur_conv = nn.Conv2d(
                cur_head_split * channels,
                cur_head_split * channels,
                kernel_size=(cur_window, cur_window),
                stride=(1, 1),
                padding=(padding_size, padding_size),
                groups=cur_head_split * channels,
            )
            self.conv_list.append(cur_conv)
            head_splits.append(cur_head_split)

        self.channel_splits = [x * channels for x in head_splits]

    def forward(self, q: torch.Tensor, v: torch.Tensor, size: tuple[int, int]) -> torch.Tensor:
        (B, L, _, C) = q.size()
        (H, W) = size
        # assert N == H * W

        v = v.reshape(B, L, H, W, C).permute(0, 1, 4, 2, 3).reshape(B, -1, H, W).contiguous()
        v_list = torch.split(v, self.channel_splits, dim=1)
        conv_v_list = []
        for i, conv in enumerate(self.conv_list):
            conv_v_list.append(conv(v_list[i]))

        conv_v = torch.concat(conv_v_list, dim=1)
        conv_v = conv_v.reshape(B, L, C, H, W).permute(0, 1, 3, 4, 2).reshape(B, L, -1, C).contiguous()

        return q * conv_v


class ConvPosEnc(nn.Module):
    def __init__(self, dim: int, kernel_size: tuple[int, int]) -> None:
        super().__init__()
        self.proj = nn.Conv2d(
            dim,
            dim,
            kernel_size=kernel_size,
            stride=(1, 1),
            padding=(kernel_size[0] // 2, kernel_size[1] // 2),
            groups=dim,
        )

    def forward(self, x: torch.Tensor, size: tuple[int, int]) -> torch.Tensor:
        (B, _, C) = x.size()
        (H, W) = size
        # assert N == H * W

        x = x.transpose(1, 2).view(B, C, H, W)
        x = self.proj(x) + x
        x = x.flatten(2).transpose(1, 2)

        return x


class EfficientAtt(nn.Module):
    def __init__(self, dim: int, num_heads: int, qkv_bias: bool, proj_drop: float) -> None:
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        self.aggregator = Aggregator(dim=dim, seg=5)

        t_dim = dim // 5 * 4
        self.crpe = ConvRelPosEnc(t_dim // num_heads, window={3: 2, 5: 3, 7: 3})

    def forward(self, x: torch.Tensor, size: tuple[int, int]) -> torch.Tensor:
        (B, N, C) = x.size()
        qkv = self.qkv(x).reshape(B, N, 3, C).permute(2, 0, 1, 3).reshape(3 * B, N, C)

        (qkv, x_agg0) = self.aggregator(qkv, size, self.num_heads)
        (q, k, v) = qkv.unbind(0)

        k_softmax = k.softmax(dim=2)
        k_softmax_t_dot_v = torch.einsum("b h n k, b h n v -> b h k v", k_softmax, v)
        eff_att = torch.einsum("b h n k, b h k v -> b h n v", q, k_softmax_t_dot_v)
        crpe = self.crpe(q, v, size=size)

        x = self.scale * eff_att + crpe
        x = x.transpose(1, 2).reshape(B, N, C // 5 * 4)
        x = torch.concat([x, x_agg0], dim=-1)

        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class PatchEmbed(nn.Module):
    def __init__(self, stride: int, in_dim: int, embedding_dim: int) -> None:
        super().__init__()
        self.stride = stride

        self.proj = SeparableConv2d(in_dim, embedding_dim, kernel_size=(3, 3), stride=(stride, stride), padding=(1, 1))
        self.norm = nn.BatchNorm2d(embedding_dim)
        self.act = nn.Hardswish()

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor, tuple[int, int]]:
        (_, _, H, W) = x.size()
        x = self.act(self.norm(self.proj(x)))
        x = x.flatten(2).transpose(1, 2)

        return (x, (H // self.stride, W // self.stride))


class GroupMixAttentionBlock(nn.Module):
    def __init__(
        self, dim: int, num_heads: int, mlp_ratio: float, qkv_bias: bool, drop: float, drop_path_rate: float
    ) -> None:
        super().__init__()
        self.cpe = ConvPosEnc(dim, kernel_size=(3, 3))
        self.norm1 = nn.LayerNorm(dim, eps=1e-6)
        self.attn = EfficientAtt(dim, num_heads=num_heads, qkv_bias=qkv_bias, proj_drop=drop)
        self.drop_path = StochasticDepth(drop_path_rate, mode="row")

        self.norm2 = nn.LayerNorm(dim, eps=1e-6)
        self.mlp = MLP(dim, [int(dim * mlp_ratio), dim], activation_layer=nn.GELU, dropout=drop)

    def forward(self, x: torch.Tensor, size: tuple[int, int]) -> torch.Tensor:
        x = self.cpe(x, size)
        x = x + self.drop_path(self.attn(self.norm1(x), size))
        x = x + self.drop_path(self.mlp(self.norm2(x)))

        return x


class GroupMixFormerStage(nn.Module):
    def __init__(
        self,
        dim: int,
        out_dim: int,
        num_heads: int,
        mlp_ratio: float,
        qkv_bias: bool,
        drop: float,
        drop_path_rate: list[float],
        depth: int,
        patch_embed_stride: int,
    ) -> None:
        super().__init__()
        self.patch_embed = PatchEmbed(
            stride=patch_embed_stride,
            in_dim=dim,
            embedding_dim=out_dim,
        )

        self.gma_blocks = nn.ModuleList()
        for i in range(depth):
            self.gma_blocks.append(
                GroupMixAttentionBlock(
                    dim=out_dim,
                    num_heads=num_heads,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=qkv_bias,
                    drop=drop,
                    drop_path_rate=drop_path_rate[i],
                )
            )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (x, size) = self.patch_embed(x)
        for block in self.gma_blocks:
            x = block(x, size)

        x = x.reshape(x.size(0), size[0], size[1], -1).permute(0, 3, 1, 2)

        return x


class GroupMixFormer(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        num_heads = 8
        qkv_bias = True
        dropout = 0.0
        embed_dims: list[int] = self.config["embed_dims"]
        depths: list[int] = self.config["depths"]
        mlp_ratios: list[float] = self.config["mlp_ratios"]
        drop_path_rate: float = self.config["drop_path_rate"]

        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(depths)).split(depths)]
        num_stages = len(depths)

        self.stem = nn.Sequential(
            Conv2dNormActivation(
                self.input_channels,
                embed_dims[0] // 2,
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                activation_layer=nn.Hardswish,
                bias=True,
            ),
            Conv2dNormActivation(
                embed_dims[0] // 2,
                embed_dims[0],
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                activation_layer=nn.Hardswish,
                bias=True,
            ),
        )

        prev_dim = embed_dims[0]
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i in range(num_stages):
            stages[f"stage{i+1}"] = GroupMixFormerStage(
                prev_dim,
                embed_dims[i],
                num_heads=num_heads,
                mlp_ratio=mlp_ratios[i],
                qkv_bias=qkv_bias,
                drop=dropout,
                drop_path_rate=dpr[i],
                depth=depths[i],
                patch_embed_stride=1 if i == 0 else 2,
            )
            return_channels.append(embed_dims[i])
            prev_dim = embed_dims[i]

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.BatchNorm2d(embed_dims[-1]),
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = embed_dims[-1]
        self.classifier = self.create_classifier()

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, nn.LayerNorm):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config(
    "groupmixformer_mobile",
    GroupMixFormer,
    config={
        "embed_dims": [40, 80, 160, 160],
        "depths": [3, 3, 12, 4],
        "mlp_ratios": [4.0, 4.0, 4.0, 4.0],
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "groupmixformer_t",
    GroupMixFormer,
    config={
        "embed_dims": [80, 160, 200, 240],
        "depths": [4, 4, 12, 4],
        "mlp_ratios": [4.0, 4.0, 4.0, 4.0],
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "groupmixformer_s",
    GroupMixFormer,
    config={
        "embed_dims": [80, 160, 320, 320],
        "depths": [2, 4, 12, 4],
        "mlp_ratios": [4.0, 4.0, 4.0, 4.0],
        "drop_path_rate": 0.2,
    },
)
registry.register_model_config(
    "groupmixformer_b",
    GroupMixFormer,
    config={
        "embed_dims": [200, 240, 320, 480],
        "depths": [8, 8, 12, 8],
        "mlp_ratios": [2.0, 2.0, 4.0, 4.0],
        "drop_path_rate": 0.4,
    },
)
registry.register_model_config(
    "groupmixformer_l",
    GroupMixFormer,
    config={
        "embed_dims": [240, 320, 360, 480],
        "depths": [8, 10, 30, 10],
        "mlp_ratios": [4.0, 4.0, 2.0, 2.0],
        "drop_path_rate": 0.5,
    },
)

registry.register_weights(
    "groupmixformer_mobile_il-common",
    {
        "description": "GroupMixFormer mobile model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 22.0,
                "sha256": "c2f0607fe1691b428ed5f15b0a8ecc6333d78c653be736fcd5a6fb576642b417",
            }
        },
        "net": {"network": "groupmixformer_mobile", "tag": "il-common"},
    },
)



"""
EfficientFormer v1, adapted from
https://github.com/snap-research/EfficientFormer/blob/main/models/efficientformer.py
and
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/efficientformer.py

Paper "EfficientFormer: Vision Transformers at MobileNet Speed",
https://arxiv.org/abs/2206.01191

Changes from original:
* Removed attention bias cache
"""

# Reference license: Apache-2.0 (both)

from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import Permute
from torchvision.ops import StochasticDepth

from birder.layers import LayerScale
from birder.layers import LayerScale2d
from birder.model_registry import registry
from birder.net.base import BaseNet
from birder.net.base import interpolate_attention_bias


class Attention(nn.Module):
    def __init__(self, dim: int, key_dim: int, num_heads: int, attn_ratio: float, resolution: tuple[int, int]) -> None:
        super().__init__()
        self.num_heads = num_heads
        self.scale = key_dim**-0.5
        self.key_dim = key_dim
        self.val_dim = int(attn_ratio * key_dim)
        self.val_attn_dim = self.val_dim * num_heads
        key_attn_dim = key_dim * num_heads

        self.qkv = nn.Linear(dim, key_attn_dim * 2 + self.val_attn_dim)
        self.proj = nn.Linear(self.val_attn_dim, dim)

        pos = torch.stack(
            torch.meshgrid(torch.arange(resolution[0]), torch.arange(resolution[1]), indexing="ij")
        ).flatten(1)
        rel_pos = (pos[..., :, None] - pos[..., None, :]).abs()
        rel_pos = (rel_pos[0] * resolution[1]) + rel_pos[1]
        self.attention_biases = nn.Parameter(torch.zeros(num_heads, resolution[0] * resolution[1]))
        self.attention_bias_idxs = nn.Buffer(rel_pos)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, N, _) = x.shape
        qkv = self.qkv(x)
        qkv = qkv.reshape(B, N, self.num_heads, -1).permute(0, 2, 1, 3)
        q, k, v = qkv.split([self.key_dim, self.key_dim, self.val_dim], dim=3)

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn + self.attention_biases[:, self.attention_bias_idxs]

        attn = attn.softmax(dim=-1)
        x = (attn @ v).transpose(1, 2).reshape(B, N, self.val_attn_dim)
        x = self.proj(x)

        return x


class Downsample(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
    ) -> None:
        super().__init__()
        padding = (kernel_size[0] // 2, kernel_size[1] // 2)
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=kernel_size, stride=stride, padding=padding)
        self.norm = nn.BatchNorm2d(out_channels)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv(x)
        x = self.norm(x)
        return x


class Pooling(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.pool = nn.AvgPool2d(kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), count_include_pad=False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.pool(x) - x


class ConvMLP(nn.Module):
    def __init__(self, in_features: int, hidden_features: int, drop: float) -> None:
        super().__init__()
        self.fc1 = nn.Conv2d(in_features, hidden_features, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.act = nn.GELU()
        self.fc2 = nn.Conv2d(hidden_features, in_features, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.drop = nn.Dropout(drop)
        self.norm1 = nn.BatchNorm2d(hidden_features)
        self.norm2 = nn.BatchNorm2d(in_features)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.fc1(x)
        x = self.norm1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.norm2(x)
        x = self.drop(x)

        return x


class MetaBlock1d(nn.Module):
    def __init__(
        self,
        dim: int,
        mlp_ratio: float,
        resolution: tuple[int, int],
        proj_drop: float,
        drop_path: float,
        layer_scale_init_value: float,
    ) -> None:
        super().__init__()
        self.norm1 = nn.LayerNorm(dim)
        self.token_mixer = Attention(dim, key_dim=32, num_heads=8, attn_ratio=4.0, resolution=resolution)
        self.norm2 = nn.LayerNorm(dim)
        self.mlp = MLP(dim, [int(dim * mlp_ratio), dim], activation_layer=nn.GELU, dropout=proj_drop)

        self.drop_path = StochasticDepth(drop_path, mode="row")
        self.ls1 = LayerScale(dim, layer_scale_init_value)
        self.ls2 = LayerScale(dim, layer_scale_init_value)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.drop_path(self.ls1(self.token_mixer(self.norm1(x))))
        x = x + self.drop_path(self.ls2(self.mlp(self.norm2(x))))
        return x


class MetaBlock2d(nn.Module):
    def __init__(
        self,
        dim: int,
        mlp_ratio: float,
        proj_drop: float,
        drop_path: float,
        layer_scale_init_value: float,
    ):
        super().__init__()
        self.token_mixer = Pooling()
        self.ls1 = LayerScale2d(dim, layer_scale_init_value)
        self.drop_path = StochasticDepth(drop_path, mode="row")

        self.mlp = ConvMLP(dim, hidden_features=int(dim * mlp_ratio), drop=proj_drop)
        self.ls2 = LayerScale2d(dim, layer_scale_init_value)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.drop_path(self.ls1(self.token_mixer(x)))
        x = x + self.drop_path(self.ls2(self.mlp(x)))
        return x


class Flat(nn.Module):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x.flatten(2).transpose(1, 2)


class EfficientFormerStage(nn.Module):
    def __init__(
        self,
        dim: int,
        dim_out: int,
        depth: int,
        downsample: bool,
        num_vit: int,
        mlp_ratio: float,
        resolution: tuple[int, int],
        proj_drop: float,
        drop_path: list[float],
        layer_scale_init_value: float,
    ):
        super().__init__()
        if downsample is True:
            self.downsample = Downsample(in_channels=dim, out_channels=dim_out, kernel_size=(3, 3), stride=(2, 2))
            dim = dim_out
        else:
            assert dim == dim_out
            self.downsample = nn.Identity()

        blocks = []
        if num_vit >= depth:
            blocks.append(Flat())

        for block_idx in range(depth):
            remain_idx = depth - block_idx - 1
            if num_vit > remain_idx:
                blocks.append(
                    MetaBlock1d(
                        dim,
                        mlp_ratio=mlp_ratio,
                        resolution=resolution,
                        proj_drop=proj_drop,
                        drop_path=drop_path[block_idx],
                        layer_scale_init_value=layer_scale_init_value,
                    )
                )
            else:
                blocks.append(
                    MetaBlock2d(
                        dim,
                        mlp_ratio=mlp_ratio,
                        proj_drop=proj_drop,
                        drop_path=drop_path[block_idx],
                        layer_scale_init_value=layer_scale_init_value,
                    )
                )
                if num_vit > 0 and num_vit == remain_idx:
                    blocks.append(Flat())

        self.blocks = nn.Sequential(*blocks)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.downsample(x)
        x = self.blocks(x)
        return x


# pylint: disable=invalid-name
class EfficientFormer_v1(BaseNet):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        resolution = (int(self.size[0] / (2**5)), int(self.size[1] / (2**5)))
        layer_scale_init_value = 1e-5
        embed_dims: tuple[int, int, int, int] = self.config["embed_dims"]
        depths: tuple[int, int, int, int] = (3, 2, 6, 4)
        drop_path_rate: float = 0.0
        num_vit: int = 1

        self.stem = nn.Sequential(
            Conv2dNormActivation(
                self.input_channels, embed_dims[0] // 2, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)
            ),
            Conv2dNormActivation(embed_dims[0] // 2, embed_dims[0], kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
        )

        prev_dim = embed_dims[0]
        num_stages = len(depths)
        last_stage = num_stages - 1
        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(depths)).split(depths)]
        downsample = (False,) + (True,) * (num_stages - 1)

        stages = []
        for i in range(num_stages):
            stages.append(
                EfficientFormerStage(
                    prev_dim,
                    embed_dims[i],
                    depths[i],
                    downsample=downsample[i],
                    num_vit=num_vit if i == last_stage else 0,
                    mlp_ratio=4.0,
                    resolution=resolution,
                    proj_drop=0.0,
                    drop_path=dpr[i],
                    layer_scale_init_value=layer_scale_init_value,
                )
            )
            prev_dim = embed_dims[i]

        self.body = nn.Sequential(*stages)

        self.features = nn.Sequential(
            nn.LayerNorm(embed_dims[-1], eps=1e-6),
            Permute([0, 2, 1]),
            nn.AdaptiveAvgPool1d(output_size=1),
            nn.Flatten(1),
        )
        self.embedding_size = embed_dims[-1]
        self.dist_classifier = self.create_classifier()
        self.classifier = self.create_classifier()
        self.distillation_output = False

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)

    def reset_classifier(self, num_classes: int) -> None:
        self.num_classes = num_classes
        self.dist_classifier = self.create_classifier()
        self.classifier = self.create_classifier()

    def freeze(self, freeze_classifier: bool = True, unfreeze_features: bool = False) -> None:
        for param in self.parameters():
            param.requires_grad = False

        if freeze_classifier is False:
            for param in self.classifier.parameters():
                param.requires_grad = True

            for param in self.dist_classifier.parameters():
                param.requires_grad = True

        if unfreeze_features is True:
            for param in self.features.parameters():
                param.requires_grad = True

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def set_distillation_output(self, enable: bool = True) -> None:
        self.distillation_output = enable

    def classify(self, x: torch.Tensor) -> torch.Tensor:
        x_cls = self.classifier(x)
        x_dist = self.dist_classifier(x)

        if self.training is True and self.distillation_output is True:
            x = torch.stack([x_cls, x_dist], dim=1)
        else:
            # Classifier "token" as an average of both tokens (during normal training or inference)
            x = (x_cls + x_dist) / 2

        return x

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        assert dynamic_size is False, "Dynamic size not supported for this network"

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        old_size = self.size
        super().adjust_size(new_size)

        old_resolution = (int(old_size[0] / (2**5)), int(old_size[1] / (2**5)))
        resolution = (int(new_size[0] / (2**5)), int(new_size[1] / (2**5)))
        for m in self.body.modules():
            if isinstance(m, Attention):
                m.attention_biases = nn.Parameter(
                    interpolate_attention_bias(m.attention_biases, old_resolution, resolution)
                )

                pos = torch.stack(
                    torch.meshgrid(torch.arange(resolution[0]), torch.arange(resolution[1]), indexing="ij")
                ).flatten(1)
                rel_pos = (pos[..., :, None] - pos[..., None, :]).abs()
                rel_pos = (rel_pos[0] * resolution[1]) + rel_pos[1]
                m.attention_bias_idxs = nn.Buffer(rel_pos)


registry.register_model_config(
    "efficientformer_v1_l1",
    EfficientFormer_v1,
    config={"embed_dims": (48, 96, 224, 448), "depths": (3, 2, 6, 4), "drop_path_rate": 0.0, "num_vit": 1},
)
registry.register_model_config(
    "efficientformer_v1_l3",
    EfficientFormer_v1,
    config={"embed_dims": (64, 128, 320, 512), "depths": (4, 4, 12, 6), "drop_path_rate": 0.1, "num_vit": 4},
)
registry.register_model_config(
    "efficientformer_v1_l7",
    EfficientFormer_v1,
    config={"embed_dims": (96, 192, 384, 768), "depths": (6, 6, 18, 8), "drop_path_rate": 0.1, "num_vit": 8},
)

registry.register_weights(
    "efficientformer_v1_l1_il-common",
    {
        "description": "EfficientFormer v1 L1 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 45,
                "sha256": "f0497d4e7bad7035ca84da57601e52ce303126b481197a916520b3761f0bd652",
            }
        },
        "net": {"network": "efficientformer_v1_l1", "tag": "il-common"},
    },
)



"""
SqueezeNet v1.1, adapted from
https://github.com/pytorch/vision/blob/main/torchvision/models/squeezenet.py

Paper "SqueezeNet: AlexNet-level accuracy with 50x fewer parameters and <0.5MB model size",
https://arxiv.org/abs/1602.07360
"""

# Reference license: BSD 3-Clause

from typing import Any
from typing import Optional

import torch
from torch import nn

from birder.net.base import BaseNet


class Fire(nn.Module):
    def __init__(self, in_planes: int, squeeze: int, expand: int) -> None:
        super().__init__()
        self.squeeze = nn.Conv2d(in_planes, squeeze, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False)
        self.squeeze_activation = nn.ReLU(inplace=True)
        self.left = nn.Conv2d(squeeze, expand, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False)
        self.left_activation = nn.ReLU(inplace=True)
        self.right = nn.Conv2d(squeeze, expand, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False)
        self.right_activation = nn.ReLU(inplace=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.squeeze_activation(self.squeeze(x))
        left = self.left_activation(self.left(x))
        right = self.right_activation(self.right(x))

        x = torch.concat((left, right), dim=1)

        return x


class SqueezeNet(BaseNet):
    default_size = (227, 227)
    auto_register = True

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is None, "config not supported"

        self.stem = nn.Sequential(
            nn.Conv2d(self.input_channels, 64, kernel_size=(3, 3), stride=(2, 2), padding=(0, 0), bias=False),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=(3, 3), stride=(2, 2), padding=(0, 0), ceil_mode=True),
        )
        self.body = nn.Sequential(
            Fire(64, 16, 64),
            Fire(128, 16, 64),
            nn.MaxPool2d(kernel_size=(3, 3), stride=(2, 2), padding=(0, 0), ceil_mode=True),
            Fire(128, 32, 128),
            Fire(256, 32, 128),
            nn.MaxPool2d(kernel_size=(3, 3), stride=(2, 2), padding=(0, 0), ceil_mode=True),
            Fire(256, 48, 192),
            Fire(384, 48, 192),
            Fire(384, 64, 256),
            Fire(512, 64, 256),
        )
        self.embedding_size = 512
        self.classifier = self.create_classifier()

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_uniform_(m.weight)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        return self.forward_features(x)

    def create_classifier(self, embed_dim: Optional[int] = None) -> nn.Module:
        if self.num_classes == 0:
            return nn.Identity()

        if embed_dim is None:
            embed_dim = self.embedding_size

        return nn.Sequential(
            nn.Dropout(p=0.5, inplace=True),
            nn.Conv2d(
                embed_dim,
                self.num_classes,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
            ),
            nn.ReLU(inplace=True),
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )



"""
ShuffleNet v1, adapted from
https://github.com/megvii-model/ShuffleNet-Series/blob/master/ShuffleNetV1/network.py

Paper "ShuffleNet: An Extremely Efficient Convolutional Neural Network for Mobile Devices",
https://arxiv.org/abs/1707.01083
"""

# Reference license: MIT

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import Conv2dNormActivation

from birder.model_registry import registry
from birder.net.base import DetectorBackbone


def channel_shuffle(x: torch.Tensor, groups: int) -> torch.Tensor:
    (batch_size, num_channels, height, width) = x.size()
    channels_per_group = num_channels // groups

    # Reshape
    x = x.view(batch_size, groups, channels_per_group, height, width)

    # Transpose
    # contiguous required if transpose is used before view (https://github.com/pytorch/pytorch/issues/764)
    x = torch.transpose(x, 1, 2).contiguous()

    # Flatten
    x = x.view(batch_size, -1, height, width)

    return x


class ChannelShuffle(nn.Module):
    def __init__(self, groups: int) -> None:
        super().__init__()
        self.groups = groups

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return channel_shuffle(x, groups=self.groups)


class ShuffleUnit(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, groups: int, grouped_conv: bool) -> None:
        super().__init__()
        assert in_channels <= out_channels

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.groups = groups
        self.grouped_conv = grouped_conv
        bottleneck_channels = out_channels // 4

        if in_channels == out_channels:
            dw_conv_stride = 1
        elif in_channels < out_channels:
            dw_conv_stride = 2
            out_channels -= in_channels
        else:
            raise ValueError("in_channels must be smaller or equal to out_channels")

        if grouped_conv is True:
            first_groups = groups
        else:
            first_groups = 1

        self.bottleneck = Conv2dNormActivation(
            in_channels,
            bottleneck_channels,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            groups=first_groups,
            bias=False,
        )
        self.expand = nn.Sequential(
            nn.Conv2d(
                bottleneck_channels,
                bottleneck_channels,
                kernel_size=(3, 3),
                stride=(dw_conv_stride, dw_conv_stride),
                padding=(1, 1),
                groups=bottleneck_channels,
                bias=False,
            ),
            nn.BatchNorm2d(bottleneck_channels),
            nn.Conv2d(
                bottleneck_channels,
                out_channels,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                groups=groups,
                bias=False,
            ),
            nn.BatchNorm2d(out_channels),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        residual = x

        x = self.bottleneck(x)
        if self.grouped_conv is True:
            x = channel_shuffle(x, groups=self.groups)

        x = self.expand(x)
        if self.in_channels == self.out_channels:
            x = x + residual

        else:
            residual = F.avg_pool2d(  # pylint: disable=not-callable
                residual, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)
            )
            x = torch.concat((x, residual), dim=1)

        x = F.relu(x)
        return x


# pylint: disable=invalid-name
class ShuffleNet_v1(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        groups: int = self.config["groups"]
        if groups == 1:
            stage_repeats = [3, 7, 3]
            out_channels = [24, 144, 288, 576]

        elif groups == 2:
            stage_repeats = [3, 7, 3]
            out_channels = [24, 200, 400, 800]

        elif groups == 3:
            stage_repeats = [3, 7, 3]
            out_channels = [24, 240, 480, 960]

        elif groups == 4:
            stage_repeats = [3, 7, 3]
            out_channels = [24, 272, 544, 1088]

        elif groups == 8:
            stage_repeats = [3, 7, 3]
            out_channels = [24, 384, 768, 1536]

        else:
            raise ValueError(f"groups = {groups} not supported")

        self.stem = nn.Conv2d(
            self.input_channels, out_channels[0], kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), bias=False
        )

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []

        stages["stage1"] = nn.MaxPool2d(kernel_size=(3, 3), stride=(2, 2), padding=(1, 1))
        return_channels.append(out_channels[0])
        for i, repeat in enumerate(stage_repeats):
            layers = []
            if i == 0:
                grouped_conv = False
            else:
                grouped_conv = True

            layers.append(
                ShuffleUnit(
                    out_channels[i],
                    out_channels[i + 1],
                    groups=groups,
                    grouped_conv=grouped_conv,
                )
            )
            for _ in range(repeat):
                layers.append(
                    ShuffleUnit(
                        out_channels[i + 1],
                        out_channels[i + 1],
                        groups=groups,
                        grouped_conv=True,
                    )
                )

            stages[f"stage{i+2}"] = nn.Sequential(*layers)
            return_channels.append(out_channels[i + 1])

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = out_channels[-1]
        self.classifier = self.create_classifier()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config("shufflenet_v1_1", ShuffleNet_v1, config={"groups": 1})
registry.register_model_config("shufflenet_v1_2", ShuffleNet_v1, config={"groups": 2})
registry.register_model_config("shufflenet_v1_3", ShuffleNet_v1, config={"groups": 3})
registry.register_model_config("shufflenet_v1_4", ShuffleNet_v1, config={"groups": 4})
registry.register_model_config("shufflenet_v1_8", ShuffleNet_v1, config={"groups": 8})

registry.register_weights(
    "shufflenet_v1_4_il-common",
    {
        "description": "ShuffleNet v1 (g=4) model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 5.1,
                "sha256": "b3e816c4dfe75526ff38f3b276fd3d9bcea2776261caedacab78305750b98d7e",
            }
        },
        "net": {"network": "shufflenet_v1_4", "tag": "il-common"},
    },
)



"""
Paper "YOLOv3: An Incremental Improvement", https://arxiv.org/abs/1804.02767
"""

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation

from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class ResidualBlock(nn.Module):
    def __init__(self, channels: int) -> None:
        super().__init__()
        mid_channels = channels // 2
        self.block = nn.Sequential(
            Conv2dNormActivation(
                channels, mid_channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), activation_layer=nn.LeakyReLU
            ),
            Conv2dNormActivation(
                mid_channels, channels, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), activation_layer=nn.LeakyReLU
            ),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x
        x = self.block(x)
        x = x + identity

        return x


class Darknet(DetectorBackbone):
    default_size = (256, 256)

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        filters: list[int] = self.config.get("filters", [64, 128, 256, 512, 1024])
        repeats: list[int] = self.config["repeats"]

        self.stem = nn.Sequential(
            Conv2dNormActivation(
                self.input_channels,
                32,
                kernel_size=(3, 3),
                stride=(1, 1),
                padding=(1, 1),
                activation_layer=nn.LeakyReLU,
            ),
            Conv2dNormActivation(
                32,
                filters[0],
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                activation_layer=nn.LeakyReLU,
            ),
        )

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i, num in enumerate(repeats):
            layers = []
            if i > 0:
                layers.append(
                    Conv2dNormActivation(
                        filters[i - 1],
                        filters[i],
                        kernel_size=(3, 3),
                        stride=(2, 2),
                        padding=(1, 1),
                        activation_layer=nn.LeakyReLU,
                    )
                )

            for _ in range(num):
                layers.append(ResidualBlock(filters[i]))

            stages[f"stage{i}"] = nn.Sequential(*layers)
            return_channels.append(filters[i])

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.embedding_size = filters[-1]
        self.return_channels = return_channels[1:5]
        self.classifier = self.create_classifier()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config("darknet_17", Darknet, config={"repeats": [1, 1, 1, 1, 1]})
registry.register_model_config("darknet_21", Darknet, config={"repeats": [1, 1, 1, 2, 2]})
registry.register_model_config("darknet_53", Darknet, config={"repeats": [1, 2, 8, 8, 4]})



"""
SimMIM, adapted from
https://github.com/microsoft/Swin-Transformer/blob/main/models/simmim.py

Paper "SimMIM: A Simple Framework for Masked Image Modeling",
https://arxiv.org/abs/2111.09886
"""

# Reference license: MIT

from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn

from birder.common.masking import uniform_mask
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.mim.base import MIMBaseNet


def norm_targets(targets: torch.Tensor, patch_size: int) -> torch.Tensor:
    assert patch_size % 2 == 1

    targets_count = torch.ones_like(targets)
    targets_square = targets**2.0

    targets_mean = F.avg_pool2d(  # pylint: disable=not-callable
        targets,
        kernel_size=(patch_size, patch_size),
        stride=(1, 1),
        padding=(patch_size // 2, patch_size // 2),
        count_include_pad=False,
    )
    targets_square_mean = F.avg_pool2d(  # pylint: disable=not-callable
        targets_square,
        kernel_size=(patch_size, patch_size),
        stride=(1, 1),
        padding=(patch_size // 2, patch_size // 2),
        count_include_pad=False,
    )
    targets_count = F.avg_pool2d(  # pylint: disable=not-callable
        targets_count,
        kernel_size=(patch_size, patch_size),
        stride=(1, 1),
        padding=(patch_size // 2, patch_size // 2),
        count_include_pad=True,
    ) * (patch_size**2)

    targets_var = (targets_square_mean - targets_mean**2.0) * (targets_count / (targets_count - 1))
    targets_var = torch.clamp(targets_var, min=0.0)

    return (targets - targets_mean) / (targets_var + 1.0e-6) ** 0.5


class SimMIM(MIMBaseNet):
    default_size = (192, 192)

    def __init__(
        self,
        encoder: PreTrainEncoder,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(encoder, config=config, size=size)
        assert self.config is None, "config not supported"
        assert isinstance(self.encoder, MaskedTokenRetentionMixin)

        self.mask_ratio = 0.6
        self.patch_size = encoder.max_stride

        # self.mask_token = nn.Parameter(torch.zeros(1, self.encoder.embedding_size, 1, 1))
        self.decoder = nn.Conv2d(
            in_channels=self.encoder.embedding_size,
            out_channels=self.patch_size**2 * 3,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            bias=True,
        )

        self.mask_token = nn.Parameter(torch.zeros(1, 1, 1, self.encoder.stem_width), requires_grad=True)

        # Weights initialization
        nn.init.trunc_normal_(self.mask_token, mean=0.0, std=0.02)

    def patchify(self, imgs: torch.Tensor) -> torch.Tensor:
        """
        imgs: (N, 3, H, W)
        x: (N, L, patch_size**2 * 3)
        """

        p = self.patch_size
        assert imgs.shape[2] == imgs.shape[3] and imgs.shape[2] % p == 0

        h = imgs.shape[2] // p
        w = imgs.shape[3] // p
        x = imgs.reshape(shape=(imgs.shape[0], 3, h, p, w, p))
        x = torch.einsum("nchpwq->nhwpqc", x)
        x = x.reshape(shape=(imgs.shape[0], h * w, p**2 * 3))

        return x

    def unpatchify(self, x: torch.Tensor) -> torch.Tensor:
        """
        x: (N, conv_out**2, L*3) or (N, L*3, conv_out, conv_out)
        imgs: (N, 3, H, W)
        """

        if x.ndim == 4:
            (n, c, _, _) = x.shape
            x = x.reshape(n, c, -1)
            x = torch.einsum("ncl->nlc", x)

        p = self.patch_size
        h = int(x.shape[1] ** 0.5)
        w = int(x.shape[1] ** 0.5)
        assert h * w == x.shape[1]

        x = x.reshape(shape=(x.shape[0], h, w, p, p, 3))
        x = torch.einsum("nhwpqc->nchpwq", x)
        imgs = x.reshape(shape=(x.shape[0], 3, h * p, h * p))

        return imgs

    def forward_decoder(self, x: torch.Tensor) -> torch.Tensor:
        x = self.decoder(x)
        return x

    def forward_loss(self, x: torch.Tensor, pred: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        """
        mask: 0 is keep, 1 is remove
        """

        (N, C, _, _) = pred.shape
        pred = pred.reshape(N, C, -1)
        pred = torch.einsum("ncl->nlc", pred)

        target = norm_targets(x, 47)
        target = self.patchify(target)

        loss = F.l1_loss(pred, target, reduction="none")
        loss = loss.mean(dim=-1)
        loss = (loss * mask).sum() / mask.sum()  # Mean loss on removed patches

        return loss

    def forward(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        seq_len = (self.size[0] // self.encoder.max_stride) * (self.size[1] // self.encoder.max_stride)
        mask = uniform_mask(x.size(0), seq_len, mask_ratio=self.mask_ratio, device=x.device)[0]

        latent = self.encoder.masked_encoding_retention(x, mask, mask_token=self.mask_token)
        pred = self.forward_decoder(latent["features"])
        loss = self.forward_loss(x, pred, mask)

        return {"loss": loss, "pred": pred, "mask": mask}



from typing import Literal
from typing import NotRequired
from typing import TypedDict

FileFormatType = Literal["pt", "safetensors", "pt2", "pts", "ptl"]

FormatInfoType = TypedDict(
    "FormatInfoType",
    {"file_size": float, "sha256": str},
)

NetworkInfoType = TypedDict(
    "NetworkInfoType",
    {
        "network": str,
        "tag": NotRequired[str],
        "reparameterized": NotRequired[bool],
    },
)

ModelMetadataType = TypedDict(
    "ModelMetadataType",
    {
        "url": NotRequired[str],
        "description": str,
        "resolution": tuple[int, int],
        "formats": dict[FileFormatType, FormatInfoType],
        "net": NetworkInfoType,
        "backbone": NotRequired[NetworkInfoType],
        "encoder": NotRequired[NetworkInfoType],
        "task": NotRequired[str],
    },
)

REGISTRY_MANIFEST: dict[str, ModelMetadataType] = {}



from birder.introspection.attention_rollout import AttentionRolloutInterpreter
from birder.introspection.gradcam import GradCamInterpreter
from birder.introspection.guided_backprop import GuidedBackpropInterpreter

__all__ = [
    "AttentionRolloutInterpreter",
    "GradCamInterpreter",
    "GuidedBackpropInterpreter",
]



import argparse
import json
import logging
from pathlib import Path
from typing import Any

import onnx
import onnx.checker
import torch
import torch.onnx
from torch.utils.mobile_optimizer import optimize_for_mobile

from birder.common import cli
from birder.common import fs_ops
from birder.common import lib
from birder.data.transforms.classification import RGBType
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import SignatureType
from birder.net.base import reparameterize_available
from birder.net.detection.base import DetectionSignatureType
from birder.version import __version__

logger = logging.getLogger(__name__)


def reparameterize(
    net: torch.nn.Module,
    signature: SignatureType | DetectionSignatureType,
    class_to_idx: dict[str, int],
    rgb_stats: RGBType,
    epoch: int,
    network_name: str,
) -> None:
    if reparameterize_available(net) is False:
        logger.error("Reparameterize not supported for this network")
    else:
        net.reparameterize_model()
        network_name = lib.get_network_name(network_name, tag="reparameterized")
        fs_ops.checkpoint_model(
            network_name,
            epoch,
            net,
            signature=signature,
            class_to_idx=class_to_idx,
            rgb_stats=rgb_stats,
            optimizer=None,
            scheduler=None,
            scaler=None,
            model_base=None,
        )


def pt2_export(
    net: torch.nn.Module,
    signature: SignatureType | DetectionSignatureType,
    class_to_idx: dict[str, int],
    rgb_stats: RGBType,
    device: torch.device,
    model_path: str | Path,
) -> None:
    signature["inputs"][0]["data_shape"][0] = 2  # Set batch size
    sample_shape = signature["inputs"][0]["data_shape"]
    batch_dim = torch.export.Dim("batch", min=1, max=4096)
    exported_net = torch.export.export(
        net, (torch.randn(*sample_shape, device=device),), dynamic_shapes={"x": {0: batch_dim}}
    )
    fs_ops.save_pt2(exported_net, model_path, net.task, class_to_idx, signature, rgb_stats)


def onnx_export(
    net: torch.nn.Module,
    signature: SignatureType | DetectionSignatureType,
    class_to_idx: dict[str, int],
    model_path: str | Path,
    dynamo: bool,
    trace: bool,
) -> None:
    signature["inputs"][0]["data_shape"][0] = 1  # Set batch size
    sample_shape = signature["inputs"][0]["data_shape"]

    if dynamo is False:
        if trace is False:
            scripted_module = torch.jit.script(net)
        else:
            scripted_module = net

        torch.onnx.export(
            scripted_module,
            torch.randn(sample_shape),
            str(model_path),
            export_params=True,
            opset_version=18,
            input_names=["input"],
            output_names=["output"],
            dynamic_axes={"input": {0: "batch_size"}, "output": {0: "batch_size"}},
            dynamo=False,
        )

    else:
        batch_dim = torch.export.Dim("batch", min=1, max=4096)
        torch.onnx.export(
            net,
            torch.randn(sample_shape),
            str(model_path),
            export_params=True,
            opset_version=18,
            input_names=["input"],
            output_names=["output"],
            dynamic_shapes={"x": {0: batch_dim}},
            dynamo=True,
            external_data=False,
        )

    signature["inputs"][0]["data_shape"][0] = 0

    logger.info("Saving class to index json...")
    with open(f"{model_path}_class_to_idx.json", "w", encoding="utf-8") as handle:
        json.dump(class_to_idx, handle, indent=2)

    # Test exported model
    onnx_model = onnx.load(str(model_path))
    onnx.checker.check_model(onnx_model, full_check=True)


def config_export(
    net: torch.nn.Module, signature: SignatureType | DetectionSignatureType, rgb_stats: RGBType, model_path: str | Path
) -> None:
    model_config = lib.get_network_config(net, signature, rgb_stats)
    logger.info("Saving model config json...")
    with open(f"{model_path}_config.json", "w", encoding="utf-8") as handle:
        json.dump(model_config, handle, indent=2)


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "convert-model",
        allow_abbrev=False,
        help="convert PyTorch model to various formats",
        description="convert PyTorch model to various formats",
        epilog=(
            "Usage examples:\n"
            "python -m birder.tools convert-model --network shufflenet_v2_2_0 --epoch 200 --pts\n"
            "python -m birder.tools convert-model --network squeezenet --epoch 100 --onnx\n"
            "python -m birder.tools convert-model -n mobilevit_v2_1_5 -t intermediate -e 80 --pt2\n"
            "python -m birder.tools convert-model -n efficientnet_v2_m -e 0 --lite\n"
            "python -m birder.tools convert-model --network faster_rcnn --backbone resnext_101 -e 0 --pts\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument(
        "-n", "--network", type=str, required=True, help="the neural network to load (i.e. resnet_v2_50)"
    )
    subparser.add_argument(
        "--model-config",
        action=cli.FlexibleDictAction,
        help=(
            "override the model default configuration, accepts key-value pairs or JSON "
            "('drop_path_rate=0.2' or '{\"units\": [3, 24, 36, 3], \"dropout\": 0.2}'"
        ),
    )
    subparser.add_argument("--backbone", type=str, help="the neural network to used as backbone")
    subparser.add_argument(
        "--backbone-model-config",
        action=cli.FlexibleDictAction,
        help=(
            "override the backbone default configuration, accepts key-value pairs or JSON "
            "('drop_path_rate=0.2' or '{\"units\": [3, 24, 36, 3], \"dropout\": 0.2}'"
        ),
    )
    subparser.add_argument("--backbone-tag", type=str, help="backbone training log tag (loading only)")
    subparser.add_argument("-e", "--epoch", type=int, metavar="N", help="model checkpoint to load")
    subparser.add_argument("-t", "--tag", type=str, help="model tag (from the training phase)")
    subparser.add_argument(
        "-r", "--reparameterized", default=False, action="store_true", help="load reparameterized model"
    )
    subparser.add_argument(
        "--trace",
        default=False,
        action="store_true",
        help="trace instead of script (applies only to TorchScript conversions)",
    )
    subparser.add_argument("--force", action="store_true", help="override existing model")

    format_group = subparser.add_mutually_exclusive_group(required=True)
    format_group.add_argument("--resize", type=int, nargs="+", metavar=("H", "W"), help="resize model (pt)")
    format_group.add_argument("--resize-patch", type=int, help="resize model flexivit style model patch size (pt)")
    format_group.add_argument(
        "--add-config", action=cli.FlexibleDictAction, help=("add custom configuration to existing model (pt)")
    )
    format_group.add_argument(
        "--add-backbone-config",
        action=cli.FlexibleDictAction,
        help=("add custom configuration to existing model's backbone (pt)"),
    )
    format_group.add_argument("--reparameterize", default=False, action="store_true", help="reparameterize model")
    format_group.add_argument("--pts", default=False, action="store_true", help="convert to TorchScript model")
    format_group.add_argument(
        "--lite", default=False, action="store_true", help="convert to lite TorchScript interpreter version model"
    )
    format_group.add_argument(
        "--pt2", default=False, action="store_true", help="convert to standardized model representation"
    )
    format_group.add_argument(
        "--st", "--safetensors", default=False, action="store_true", help="convert to Safetensors"
    )
    format_group.add_argument("--onnx", default=False, action="store_true", help="convert to ONNX format")
    format_group.add_argument(
        "--onnx-dynamo", default=False, action="store_true", help="convert to ONNX format using TorchDynamo"
    )
    format_group.add_argument("--config", default=False, action="store_true", help="generate model config json")

    subparser.set_defaults(func=main)


# pylint: disable=too-many-branches
def main(args: argparse.Namespace) -> None:
    args.resize = cli.parse_size(args.resize)

    if args.backbone is not None and registry.exists(args.backbone, net_type=DetectorBackbone) is False:
        raise cli.ValidationError(
            f"--backbone {args.network} not supported, see list-models tool for available options"
        )
    if args.trace is True and args.pts is False and args.lite is False and args.onnx is False:
        raise cli.ValidationError("--trace requires one of --pts, --lite --onnx to be set")

    # Load model
    device = torch.device("cpu")
    signature: SignatureType | DetectionSignatureType
    backbone_custom_config = None
    if args.backbone is None:
        (net, (class_to_idx, signature, rgb_stats, custom_config)) = fs_ops.load_model(
            device,
            args.network,
            config=args.model_config,
            tag=args.tag,
            epoch=args.epoch,
            new_size=args.resize,
            inference=True,
            reparameterized=args.reparameterized,
        )
        network_name = lib.get_network_name(args.network, tag=args.tag)

    else:
        (net, (class_to_idx, signature, rgb_stats, custom_config, backbone_custom_config)) = (
            fs_ops.load_detection_model(
                device,
                args.network,
                config=args.model_config,
                tag=args.tag,
                backbone=args.backbone,
                backbone_config=args.backbone_model_config,
                backbone_tag=args.backbone_tag,
                epoch=args.epoch,
                new_size=args.resize,
                inference=True,
                export_mode=True,
            )
        )
        network_name = lib.get_detection_network_name(
            args.network, tag=args.tag, backbone=args.backbone, backbone_tag=args.backbone_tag
        )

    if args.resize is not None:
        network_name = f"{network_name}_{args.resize[0]}px"
    elif args.add_config is not None or args.add_backbone_config is not None:
        network_name = f"{network_name}_custom_config"
    elif args.resize_patch is not None:
        network_name = f"{network_name}_ip{args.resize_patch}"

    model_path = fs_ops.model_path(
        network_name,
        epoch=args.epoch,
        pts=args.pts,
        lite=args.lite,
        pt2=args.pt2,
        st=args.st,
        onnx=args.onnx or args.onnx_dynamo,
    )
    if model_path.exists() is True and args.force is False and args.reparameterize is False and args.config is False:
        logger.warning("Converted model already exists... aborting")
        raise SystemExit(1)

    logger.info(f"Saving converted model {model_path}...")
    if args.resize is not None:
        net.adjust_size(args.resize)
        signature["inputs"][0]["data_shape"][2] = args.resize[0]
        signature["inputs"][0]["data_shape"][3] = args.resize[1]
        fs_ops.checkpoint_model(
            network_name,
            args.epoch,
            net,
            signature,
            class_to_idx,
            rgb_stats,
            optimizer=None,
            scheduler=None,
            scaler=None,
            model_base=None,
            external_config=custom_config,
            external_backbone_config=backbone_custom_config,
        )

    elif args.resize_patch is not None:
        net.adjust_patch_size(args.resize_patch)
        fs_ops.checkpoint_model(
            network_name,
            args.epoch,
            net,
            signature,
            class_to_idx,
            rgb_stats,
            optimizer=None,
            scheduler=None,
            scaler=None,
            model_base=None,
            external_config=custom_config,
            external_backbone_config=backbone_custom_config,
        )

    elif args.add_config is not None:
        if custom_config is not None:
            logger.warning("Custom config already exist, current action overrides it")

        fs_ops.checkpoint_model(
            network_name,
            args.epoch,
            net,
            signature,
            class_to_idx,
            rgb_stats,
            optimizer=None,
            scheduler=None,
            scaler=None,
            model_base=None,
            external_config=args.add_config,
        )

    elif args.add_backbone_config is not None:
        if backbone_custom_config is not None:
            logger.warning("Custom backbone config already exist, current action overrides it")

        fs_ops.checkpoint_model(
            network_name,
            args.epoch,
            net,
            signature,
            class_to_idx,
            rgb_stats,
            optimizer=None,
            scheduler=None,
            scaler=None,
            model_base=None,
            external_backbone_config=args.add_backbone_config,
        )

    elif args.reparameterize is True:
        reparameterize(net, signature, class_to_idx, rgb_stats, args.epoch, network_name)

    elif args.lite is True:
        if args.trace is True:
            sample_shape = [1] + signature["inputs"][0]["data_shape"][1:]  # C, H, W
            scripted_module = torch.jit.trace(net, example_inputs=torch.randn(sample_shape))
            optimized_scripted_module = scripted_module
        else:
            scripted_module = torch.jit.script(net)
            optimized_scripted_module = optimize_for_mobile(scripted_module)

        optimized_scripted_module._save_for_lite_interpreter(  # pylint: disable=protected-access
            str(model_path),
            _extra_files={
                "birder_version": __version__,
                "task": net.task,
                "class_to_idx": json.dumps(class_to_idx),
                "signature": json.dumps(signature),
                "rgb_stats": json.dumps(rgb_stats),
            },
        )

    elif args.pt2 is True:
        pt2_export(net, signature, class_to_idx, rgb_stats, device, model_path)

    elif args.st is True:
        fs_ops.save_st(
            net,
            str(model_path),
            net.task,
            class_to_idx,
            signature,
            rgb_stats,
            external_config=custom_config,
            external_backbone_config=backbone_custom_config,
        )

    elif args.onnx is True or args.onnx_dynamo is True:
        config_export(net, signature, rgb_stats, model_path)
        onnx_export(net, signature, class_to_idx, model_path, args.onnx_dynamo, args.trace)

    elif args.config is True:
        config_export(net, signature, rgb_stats, model_path)

    elif args.pts is True:
        if args.trace is True:
            sample_shape = [1] + signature["inputs"][0]["data_shape"][1:]  # C, H, W
            scripted_module = torch.jit.trace(net, example_inputs=torch.randn(sample_shape))
        else:
            scripted_module = torch.jit.script(net)

        fs_ops.save_pts(scripted_module, model_path, net.task, class_to_idx, signature, rgb_stats)



import argparse
import logging
import random
from pathlib import Path
from typing import Any
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
import torch
import torchvision.transforms.v2.functional as F
from torch.utils.data import DataLoader
from torch.utils.data.dataloader import default_collate
from torchvision.datasets import ImageFolder

from birder.common import cli
from birder.common import fs_ops
from birder.common import masking
from birder.common import training_cli
from birder.conf import settings
from birder.data.dataloader.webdataset import make_wds_loader
from birder.data.datasets.webdataset import make_wds_dataset
from birder.data.datasets.webdataset import prepare_wds_args
from birder.data.datasets.webdataset import wds_args_from_info
from birder.data.transforms.classification import get_mixup_cutmix
from birder.data.transforms.classification import get_rgb_stats
from birder.data.transforms.classification import inference_preset
from birder.data.transforms.classification import reverse_preset
from birder.data.transforms.classification import training_preset

logger = logging.getLogger(__name__)


# pylint: disable=too-many-locals,too-many-branches
def show_iterator(args: argparse.Namespace) -> None:
    reverse_transform = reverse_preset(get_rgb_stats("birder"))
    if args.mode == "training":
        transform = training_preset(
            args.size,
            args.aug_type,
            args.aug_level,
            get_rgb_stats("birder"),
            args.resize_min_scale,
            args.re_prob,
            args.use_grayscale,
            args.ra_num_ops,
            args.ra_magnitude,
            args.augmix_severity,
            args.simple_crop,
        )
    elif args.mode == "inference":
        transform = inference_preset(args.size, get_rgb_stats("birder"), args.center_crop, args.simple_crop)
    else:
        raise ValueError(f"Unknown mode={args.mode}")

    batch_size = 8
    if args.wds is True:
        wds_path: str | list[str]
        if args.wds_info is not None:
            (wds_path, dataset_size) = wds_args_from_info(args.wds_info, args.wds_split)
            if args.wds_size is not None:
                dataset_size = args.wds_size
        else:
            (wds_path, dataset_size) = prepare_wds_args(args.data_path, args.wds_size, torch.device("cpu"))

        dataset = make_wds_dataset(
            wds_path,
            dataset_size=dataset_size,
            shuffle=True,
            samples_names=False,
            transform=transform,
        )
        if args.wds_class_file is None:
            args.wds_class_file = Path(args.data_path).joinpath(settings.CLASS_LIST_NAME)

        class_to_idx = fs_ops.read_class_file(args.wds_class_file)

    else:
        dataset = ImageFolder(args.data_path, transform=transform)
        class_to_idx = dataset.class_to_idx

    no_iterations = 6
    if args.batch is False:
        samples = random.sample(dataset.imgs, no_iterations)
        cols = 4
        rows = 3
        for img_path, _ in samples:
            img = dataset.loader(img_path)
            fig = plt.figure(constrained_layout=True)
            grid_spec = fig.add_gridspec(ncols=cols, nrows=rows)

            # Show original
            ax = fig.add_subplot(grid_spec[0, 0:cols])
            ax.imshow(img)
            ax.set_title(f"Original, aug type: {args.aug_type}")

            # Show transformed
            counter = 0
            for i in range(cols):
                for j in range(1, rows):
                    transformed_img = F.to_pil_image(reverse_transform(transform(img)))

                    ax = fig.add_subplot(grid_spec[j, i])
                    ax.imshow(np.asarray(transformed_img))
                    ax.set_title(f"#{counter}")
                    counter += 1

            plt.show()

    else:
        cols = 4
        rows = 2
        num_outputs = len(class_to_idx)
        t = get_mixup_cutmix(args.mixup_alpha, num_outputs, args.cutmix)

        def collate_fn(batch: Any) -> Any:
            return t(*default_collate(batch))

        if args.wds is True:
            data_loader = make_wds_loader(
                dataset,
                batch_size,
                num_workers=1,
                prefetch_factor=1,
                collate_fn=collate_fn,
                world_size=1,
                pin_memory=False,
                shuffle=args.wds_extra_shuffle,
            )

        else:
            data_loader = DataLoader(
                dataset,
                batch_size=batch_size,
                shuffle=True,
                collate_fn=collate_fn,
            )

        # Masking
        mask_size = (args.size[0] // args.patch_size, args.size[1] // args.patch_size)
        mask_generator: Optional[masking.Masking]
        if args.masking == "uniform":
            mask_generator = masking.UniformMasking(mask_size, args.mask_ratio)
        elif args.masking == "block":
            max_patches = int(args.mask_ratio * mask_size[0] * mask_size[1])
            mask_generator = masking.BlockMasking(mask_size, 4, max_patches, 0.33, 3.33)
        else:
            mask_generator = None

        for k, (inputs, _) in enumerate(data_loader):
            if k >= no_iterations:
                break

            fig = plt.figure(constrained_layout=True)
            grid_spec = fig.add_gridspec(ncols=cols, nrows=rows)

            if mask_generator is not None:
                masks = mask_generator(batch_size)
                inputs = masking.mask_tensor(inputs, masks, patch_factor=args.patch_size)

            # Show transformed
            counter = 0
            for i in range(cols):
                for j in range(rows):
                    img = inputs[i + cols * j]
                    transformed_img = F.to_pil_image(reverse_transform(img))

                    ax = fig.add_subplot(grid_spec[j, i])
                    ax.imshow(np.asarray(transformed_img))
                    ax.set_title(f"#{counter}")
                    counter += 1

            plt.show()


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "show-iterator",
        allow_abbrev=False,
        help="show training / inference iterator output vs input",
        description="show training / inference iterator output vs input",
        epilog=(
            "Usage examples:\n"
            "python -m birder.tools show-iterator --mode training --aug-level 3\n"
            "python -m birder.tools show-iterator --mode training --size 224 --aug-level 2 --batch\n"
            "python -m birder.tools show-iterator --mode inference --size 320\n"
            "python -m birder.tools show-iterator --mode training --size 224 --batch --wds "
            "--wds-class-file ~/Datasets/imagenet-1k-wds/classes.txt --wds-size 50000 "
            "--data-path ~/Datasets/imagenet-1k-wds/validation\n"
            "python -m birder.tools show-iterator --mode training --batch --size 224 --aug-level 6 --masking uniform\n"
            "python -m birder.tools show-iterator --mode training --size 224 --batch --wds "
            "--data-path data/training_packed\n"
            "python -m birder.tools show-iterator --mode training --batch --mixup-alpha 0.8 --cutmix "
            "--aug-level 8 --data-path ~/Datasets/inat2021/train\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument(
        "--mode", type=str, choices=["training", "inference"], default="training", help="iterator mode"
    )
    subparser.add_argument("--size", type=int, nargs="+", default=[224], metavar=("H", "W"), help="image size")
    training_cli.add_data_aug_args(subparser)
    subparser.add_argument("--center-crop", type=float, default=1.0, help="center crop ratio during inference")
    subparser.add_argument(
        "--batch", default=False, action="store_true", help="show a batch instead of a single sample"
    )
    subparser.add_argument("--mixup-alpha", type=float, help="mixup alpha")
    subparser.add_argument("--cutmix", default=False, action="store_true", help="enable cutmix")
    subparser.add_argument("--masking", type=str, choices=["uniform", "block"], help="enable masking")
    subparser.add_argument("--mask-ratio", type=float, default=0.5, help="mask ratio")
    subparser.add_argument("--patch-size", type=int, default=16, help="mask base patch size")
    subparser.add_argument(
        "--data-path", type=str, default=str(settings.TRAINING_DATA_PATH), help="image directory path"
    )
    subparser.add_argument("--wds", default=False, action="store_true", help="use webdataset")
    subparser.add_argument("--wds-info", type=str, metavar="FILE", help="wds info file path")
    subparser.add_argument("--wds-class-file", type=str, metavar="FILE", help="class list file")
    subparser.add_argument("--wds-size", type=int, metavar="N", help="size of the wds directory")
    subparser.add_argument(
        "--wds-split", type=str, default="training", metavar="NAME", help="wds dataset split to load"
    )
    subparser.add_argument(
        "--wds-extra-shuffle",
        default=False,
        action="store_true",
        help="enable cross-worker batch shuffling after batching",
    )
    subparser.set_defaults(func=main)


def main(args: argparse.Namespace) -> None:
    if args.wds is True and args.batch is False:
        raise cli.ValidationError("--wds requires --batch to be set")
    if args.masking is not None and args.batch is False:
        raise cli.ValidationError("--wds requires --batch to be set")

    args.size = cli.parse_size(args.size)
    show_iterator(args)



"""
GhostNet v2, adapted from
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/ghostnet.py

Paper "GhostNetV2: Enhance Cheap Operation with Long-Range Attention", https://arxiv.org/abs/2211.12905
"""

# Reference license: Apache-2.0

import math
from collections import OrderedDict
from typing import Any
from typing import Literal
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import SqueezeExcitation

from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import make_divisible
from birder.net.ghostnet_v1 import GhostModule


# pylint: disable=invalid-name
class GhostModule_v2(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        ratio: int,
        dw_size: tuple[int, int],
        use_act: bool,
    ) -> None:
        super().__init__()
        self.gate_fn = nn.Sigmoid()
        self.out_channels = out_channels
        init_channels = math.ceil(out_channels / ratio)
        new_channels = init_channels * (ratio - 1)

        activation_layer = None
        if use_act is True:
            activation_layer = nn.ReLU

        self.primary_conv = Conv2dNormActivation(
            in_channels,
            init_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=((kernel_size[0] - 1) // 2, (kernel_size[1] - 1) // 2),
            bias=False,
            activation_layer=activation_layer,
        )

        self.cheap_operation = Conv2dNormActivation(
            init_channels,
            new_channels,
            kernel_size=dw_size,
            stride=(1, 1),
            padding=((dw_size[0] - 1) // 2, (dw_size[1] - 1) // 2),
            groups=init_channels,
            bias=False,
            activation_layer=activation_layer,
        )
        self.avg_pool = nn.AvgPool2d(kernel_size=(2, 2), stride=(2, 2), padding=(0, 0))
        self.short_conv = nn.Sequential(
            nn.Conv2d(
                in_channels,
                out_channels,
                kernel_size=kernel_size,
                stride=stride,
                padding=((kernel_size[0] - 1) // 2, (kernel_size[1] - 1) // 2),
                bias=False,
            ),
            nn.BatchNorm2d(out_channels),
            nn.Conv2d(
                out_channels,
                out_channels,
                kernel_size=(1, 5),
                stride=(1, 1),
                padding=(0, 2),
                groups=out_channels,
                bias=False,
            ),
            nn.BatchNorm2d(out_channels),
            nn.Conv2d(
                out_channels,
                out_channels,
                kernel_size=(5, 1),
                stride=(1, 1),
                padding=(2, 0),
                groups=out_channels,
                bias=False,
            ),
            nn.BatchNorm2d(out_channels),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        res = self.short_conv(self.avg_pool(x))
        x1 = self.primary_conv(x)
        x2 = self.cheap_operation(x1)
        out = torch.concat([x1, x2], dim=1)
        out = out[:, : self.out_channels, :, :] * F.interpolate(
            self.gate_fn(res), size=(out.shape[-2], out.shape[-1]), mode="nearest"
        )

        return out


class GhostBottleneck(nn.Module):
    def __init__(
        self,
        in_channels: int,
        mid_channels: int,
        out_channels: int,
        dw_kernel_size: tuple[int, int],
        stride: tuple[int, int],
        se_ratio: float,
        module_version: Literal["v1", "v2"],
    ) -> None:
        super().__init__()

        # Point-wise expansion
        if module_version == "v1":
            self.ghost1 = GhostModule(
                in_channels, mid_channels, kernel_size=(1, 1), ratio=2, dw_size=(3, 3), stride=(1, 1), use_act=True
            )

        else:
            self.ghost1 = GhostModule_v2(
                in_channels, mid_channels, kernel_size=(1, 1), ratio=2, dw_size=(3, 3), stride=(1, 1), use_act=True
            )

        # Depth-wise convolution
        if stride[0] > 1 or stride[1] > 1:
            self.conv_dw = nn.Conv2d(
                mid_channels,
                mid_channels,
                kernel_size=dw_kernel_size,
                stride=stride,
                padding=((dw_kernel_size[0] - 1) // 2, (dw_kernel_size[1] - 1) // 2),
                groups=mid_channels,
                bias=False,
            )
            self.bn_dw = nn.BatchNorm2d(mid_channels)

        else:
            self.conv_dw = nn.Identity()
            self.bn_dw = nn.Identity()

        # Squeeze-and-excitation
        if se_ratio > 0:
            squeeze_channels = make_divisible(se_ratio * mid_channels, 4)
            self.se = SqueezeExcitation(mid_channels, squeeze_channels, scale_activation=nn.Hardsigmoid)

        else:
            self.se = nn.Identity()

        # Point-wise linear projection
        self.ghost2 = GhostModule(
            mid_channels, out_channels, kernel_size=(1, 1), ratio=2, dw_size=(3, 3), stride=(1, 1), use_act=False
        )

        # shortcut
        if in_channels == out_channels and stride[0] == 1 and stride[1] == 1:
            self.shortcut = nn.Identity()

        else:
            self.shortcut = nn.Sequential(
                nn.Conv2d(
                    in_channels,
                    in_channels,
                    kernel_size=dw_kernel_size,
                    stride=stride,
                    padding=((dw_kernel_size[0] - 1) // 2, (dw_kernel_size[1] - 1) // 2),
                    groups=in_channels,
                    bias=False,
                ),
                nn.BatchNorm2d(in_channels),
                nn.Conv2d(in_channels, out_channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False),
                nn.BatchNorm2d(out_channels),
            )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        shortcut = x
        x = self.ghost1(x)
        x = self.conv_dw(x)
        x = self.bn_dw(x)
        x = self.se(x)
        x = self.ghost2(x)
        x += self.shortcut(shortcut)

        return x


# pylint: disable=invalid-name
class GhostNet_v2(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        width: float = self.config["width"]

        block_config: list[list[tuple[int, int, int, float, int]]] = [
            # kernel, expansion, channels, se ratio, stride
            # Stage 1
            [(3, 16, 16, 0, 1)],
            # Stage 2
            [(3, 48, 24, 0, 2)],
            [(3, 72, 24, 0, 1)],
            # Stage 3
            [(5, 72, 40, 0.25, 2)],
            [(5, 120, 40, 0.25, 1)],
            # Stage 4
            [(3, 240, 80, 0, 2)],
            [
                (3, 200, 80, 0, 1),
                (3, 184, 80, 0, 1),
                (3, 184, 80, 0, 1),
                (3, 480, 112, 0.25, 1),
                (3, 672, 112, 0.25, 1),
            ],
            # Stage 5
            [(5, 672, 160, 0.25, 2)],
            [
                (5, 960, 160, 0, 1),
                (5, 960, 160, 0.25, 1),
                (5, 960, 160, 0, 1),
                (5, 960, 160, 0.25, 1),
            ],
        ]

        stem_channels = make_divisible(16 * width, 4)
        self.stem = Conv2dNormActivation(
            self.input_channels, stem_channels, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), bias=False
        )

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        i = 0
        layers: list[nn.Module] = []
        prev_channels = stem_channels
        layer_idx = 0
        for cfg in block_config:
            for k, exp_size, c, se_ratio, s in cfg:
                if s > 1:
                    stages[f"stage{i}"] = nn.Sequential(*layers)
                    return_channels.append(prev_channels)
                    layers = []
                    i += 1

                out_channels = make_divisible(c * width, 4)
                mid_channels = make_divisible(exp_size * width, 4)
                if layer_idx > 1:
                    module_version: Literal["v1", "v2"] = "v2"

                else:
                    module_version = "v1"

                layers.append(
                    GhostBottleneck(
                        prev_channels,
                        mid_channels,
                        out_channels,
                        (k, k),
                        (s, s),
                        se_ratio=se_ratio,
                        module_version=module_version,
                    )
                )
                prev_channels = out_channels
                layer_idx += 1

        out_channels = make_divisible(exp_size * width, 4)
        layers.append(
            Conv2dNormActivation(prev_channels, out_channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        )
        stages[f"stage{i}"] = nn.Sequential(*layers)
        return_channels.append(out_channels)

        self.body = nn.Sequential(stages)
        prev_channels = out_channels
        out_channels = 1280
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Conv2d(prev_channels, out_channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=True),
            nn.ReLU(inplace=True),
            nn.Flatten(1),
            nn.Dropout(p=0.2),
        )
        self.return_channels = return_channels[1:5]
        self.embedding_size = out_channels
        self.classifier = self.create_classifier()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config("ghostnet_v2_1_0", GhostNet_v2, config={"width": 1.0})
registry.register_model_config("ghostnet_v2_1_3", GhostNet_v2, config={"width": 1.3})
registry.register_model_config("ghostnet_v2_1_6", GhostNet_v2, config={"width": 1.6})

registry.register_weights(
    "ghostnet_v2_1_0_il-common",
    {
        "description": "GhostNet v2 1.0x model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 20.9,
                "sha256": "efb438440a88b7e159b36bbba7de1d0748c8caae16dbe85b0330ccf16ade93c0",
            }
        },
        "net": {"network": "ghostnet_v2_1_0", "tag": "il-common"},
    },
)



from typing import Any
from typing import Optional
from typing import TypedDict

import torch
from torch import nn

from birder.model_registry import Task
from birder.model_registry import registry
from birder.net.base import DataShapeType
from birder.net.base import PreTrainEncoder

MIMSignatureType = TypedDict(
    "MIMSignatureType",
    {
        "inputs": list[DataShapeType],
        "outputs": list[DataShapeType],
    },
)


def get_mim_signature(input_shape: tuple[int, ...]) -> MIMSignatureType:
    return {
        "inputs": [{"data_shape": [0, *input_shape[1:]]}],
        "outputs": [{"data_shape": [0, *input_shape[1:]]}],
    }


class MIMBaseNet(nn.Module):
    default_size: tuple[int, int]
    task = str(Task.MASKED_IMAGE_MODELING)

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        registry.register_model(cls.__name__.lower(), cls)

    def __init__(
        self,
        encoder: PreTrainEncoder,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__()
        self.input_channels = encoder.input_channels
        self.encoder = encoder
        if hasattr(self, "config") is False:  # Avoid overriding aliases
            self.config = config
        elif config is not None:
            assert self.config is not None
            self.config.update(config)  # Override with custom config

        if size is not None:
            self.size = size
        else:
            self.size = self.default_size

        assert isinstance(self.size, tuple)
        assert isinstance(self.size[0], int)
        assert isinstance(self.size[1], int)

    def forward(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        raise NotImplementedError
